# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add event id column

Revision ID: 5f183546b39c
Revises: e0eda5d09eae
Create Date: 2019-12-05 13:33:30.155447

"""

# revision identifiers, used by Alembic.
revision = '5f183546b39c'
down_revision = 'e0eda5d09eae'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade(table_prefix=''):
    op.add_column(
        table_prefix + "zuul_buildset",
        sa.Column("event_id", sa.String(255), nullable=True)
    )


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import grpc
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import \
    OTLPSpanExporter as GRPCExporter
from opentelemetry.exporter.otlp.proto.http.trace_exporter import \
    OTLPSpanExporter as HTTPExporter
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider, Span
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry import trace
from opentelemetry.sdk import trace as trace_sdk

from zuul.lib.config import get_default, any_to_bool


class ZuulSpan(Span):
    """An implementation of Span which accepts floating point
    times and converts them to the expected nanoseconds."""

    def start(self, start_time=None, parent_context=None):
        if isinstance(start_time, float):
            start_time = int(start_time * (10**9))
        return super().start(start_time, parent_context)

    def end(self, end_time=None):
        if isinstance(end_time, float):
            end_time = int(end_time * (10**9))
        return super().end(end_time)


# Patch the OpenTelemetry SDK Span class to return a ZuulSpan so that
# we can supply floating point timestamps.
trace_sdk._Span = ZuulSpan


def _formatContext(context):
    return {
        'trace_id': context.trace_id,
        'span_id': context.span_id,
    }


def _formatAttributes(attrs):
    if attrs is None:
        return None
    return attrs.copy()


def getSpanInfo(span, include_attributes=False):
    """Return a dict for use in serializing a Span."""
    links = [{'context': _formatContext(l.context),
              'is_remote': l.context.is_remote,
              'attributes': _formatAttributes(l.attributes)}
             for l in span.links]
    attrs = _formatAttributes(span.attributes)
    context = span.get_span_context()
    parent_context = None
    if span.parent:
        parent_context = {
            **_formatContext(span.parent),
            "is_remote": span.parent.is_remote,
        }
    ret = {
        'name': span.name,
        'trace_id': context.trace_id,
        'span_id': context.span_id,
        'trace_flags': context.trace_flags,
        'start_time': span.start_time,
        'parent': parent_context,
    }
    if links:
        ret['links'] = links
    if attrs:
        if not include_attributes:
            # Avoid setting attributes when we start saved spans
            # because we have to store them in ZooKeeper and we should
            # minimize what we store there (especially since it is
            # usually duplicative).  If you really need to set
            # attributes at the start of a span (because the info is
            # not available later), set include_attributes to True.
            # Otherwise, we raise an error here to remind ourselves to
            # avoid that programming pattern.
            raise RuntimeError("Attributes were set on a saved span; "
                               "either set them when ending the span, "
                               "or set include_attributes=True")
        ret['attributes'] = attrs
    return ret


def restoreSpan(span_info, is_remote=True):
    """Restore a Span from the serialized dict provided by getSpanInfo

    Return None if unable to serialize the span.
    """
    tracer = trace.get_tracer("zuul")
    if span_info is None:
        return trace.INVALID_SPAN
    required_keys = {'name', 'trace_id', 'span_id', 'trace_flags'}
    if not required_keys <= set(span_info.keys()):
        return trace.INVALID_SPAN
    span_context = trace.SpanContext(
        span_info['trace_id'],
        span_info['span_id'],
        is_remote=is_remote,
        trace_flags=trace.TraceFlags(span_info['trace_flags']),
    )
    links = []
    for link_info in span_info.get('links', []):
        link_context = trace.SpanContext(
            link_info['context']['trace_id'],
            link_info['context']['span_id'],
            is_remote=link_info['is_remote'])
        link = trace.Link(link_context, link_info['attributes'])
        links.append(link)
    attributes = span_info.get('attributes', {})
    parent_context = None
    if parent_info := span_info.get("parent"):
        parent_context = trace.SpanContext(
            parent_info['trace_id'],
            parent_info['span_id'],
            is_remote=parent_info['is_remote'],
        )

    span = ZuulSpan(
        name=span_info['name'],
        context=span_context,
        parent=parent_context,
        sampler=tracer.sampler,
        resource=tracer.resource,
        attributes=attributes,
        span_processor=tracer.span_processor,
        kind=trace.SpanKind.INTERNAL,
        links=links,
        instrumentation_info=tracer.instrumentation_info,
        record_exception=False,
        set_status_on_exception=True,
        limits=tracer._span_limits,
        instrumentation_scope=tracer._instrumentation_scope,
    )
    span.start(start_time=span_info['start_time'])

    return span


def startSavedSpan(*args, **kw):
    """Start a span and serialize it

    This is a convenience method which starts a span (either root
    or child) and immediately serializes it.

    Most spans in Zuul should use this method.
    """
    tracer = trace.get_tracer("zuul")
    include_attributes = kw.pop('include_attributes', False)
    span = tracer.start_span(*args, **kw)
    return getSpanInfo(span, include_attributes)


def endSavedSpan(span_info, end_time=None, attributes=None):
    """End a saved span.

    This is a convenience method to restore a saved span and
    immediately end it.

    Most spans in Zuul should use this method.
    """
    span = restoreSpan(span_info, is_remote=False)
    if span:
        if attributes:
            span.set_attributes(attributes)
        span.end(end_time=end_time)


def getSpanContext(span):
    """Return a dict for use in serializing a Span Context.

    The span context information used here is a lightweight
    encoding of the span information so that remote child spans
    can be started without access to a fully restored parent.
    This is equivalent to (but not the same format) as the
    OpenTelemetry trace context propogator.
    """
    context = span.get_span_context()
    return {
        'trace_id': context.trace_id,
        'span_id': context.span_id,
        'trace_flags': context.trace_flags,
    }


def restoreSpanContext(span_context):
    """Return a span with remote context information from getSpanContext.

    This returns a non-recording span to use as a parent span.  It
    avoids the necessity of fully restoring the parent span.
    """
    if span_context:
        span_context = trace.SpanContext(
            trace_id=span_context['trace_id'],
            span_id=span_context['span_id'],
            is_remote=True,
            trace_flags=trace.TraceFlags(span_context['trace_flags'])
        )
    else:
        span_context = trace.INVALID_SPAN_CONTEXT
    parent = trace.NonRecordingSpan(span_context)
    return parent


def startSpanInContext(span_context, *args, **kw):
    """Start a span in a saved context.

    This restores a span from a saved context and starts a new child span.
    """
    tracer = trace.get_tracer("zuul")
    parent = restoreSpanContext(span_context)
    with trace.use_span(parent):
        return tracer.start_span(*args, **kw)


class Tracing:
    PROTOCOL_GRPC = 'grpc'
    PROTOCOL_HTTP_PROTOBUF = 'http/protobuf'
    processor_class = BatchSpanProcessor

    def __init__(self, config):
        service_name = get_default(config, "tracing", "service_name", "zuul")
        resource = Resource(attributes={SERVICE_NAME: service_name})
        provider = TracerProvider(resource=resource)
        trace.set_tracer_provider(provider)
        enabled = get_default(config, "tracing", "enabled")
        if not any_to_bool(enabled):
            self.processor = None
            return

        protocol = get_default(config, "tracing", "protocol",
                               self.PROTOCOL_GRPC)
        endpoint = get_default(config, "tracing", "endpoint")
        tls_key = get_default(config, "tracing", "tls_key")
        tls_cert = get_default(config, "tracing", "tls_cert")
        tls_ca = get_default(config, "tracing", "tls_ca")
        certificate_file = get_default(config, "tracing", "certificate_file")
        insecure = get_default(config, "tracing", "insecure")
        if insecure is not None:
            insecure = any_to_bool(insecure)
        timeout = get_default(config, "tracing", "timeout")
        if timeout is not None:
            timeout = int(timeout)
        compression = get_default(config, "tracing", "compression")

        if protocol == self.PROTOCOL_GRPC:
            if certificate_file:
                raise Exception("The certificate_file tracing option "
                                f"is not valid for {protocol} endpoints")
            if any([tls_ca, tls_key, tls_cert]):
                if tls_ca:
                    tls_ca = open(tls_ca, 'rb').read()
                if tls_key:
                    tls_key = open(tls_key, 'rb').read()
                if tls_cert:
                    tls_cert = open(tls_cert, 'rb').read()
                creds = grpc.ssl_channel_credentials(
                    root_certificates=tls_ca,
                    private_key=tls_key,
                    certificate_chain=tls_cert)
            else:
                creds = None
            exporter = GRPCExporter(
                endpoint=endpoint,
                insecure=insecure,
                credentials=creds,
                timeout=timeout,
                compression=compression)
        elif protocol == self.PROTOCOL_HTTP_PROTOBUF:
            if insecure:
                raise Exception("The insecure tracing option "
                                f"is not valid for {protocol} endpoints")
            if any([tls_ca, tls_key, tls_cert]):
                raise Exception("The tls_* tracing options "
                                f"are not valid for {protocol} endpoints")
            exporter = HTTPExporter(
                endpoint=endpoint,
                certificate_file=certificate_file,
                timeout=timeout,
                compression=compression)
        else:
            raise Exception(f"Unknown tracing protocol {protocol}")
        self.processor = self.processor_class(exporter)
        provider.add_span_processor(self.processor)

    def stop(self):
        if not self.processor:
            return
        self.processor.shutdown()



# Copyright (c) 2017 Red Hat
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.

import os
import json
import tempfile
from copy import deepcopy

from ansible.plugins.action import ActionBase
from zuul.ansible.schema import (
    artifact_schema,
    warning_schema,
)


def validate_schema(data):
    artifact_schema(data)
    warning_schema(data)


def merge_dict(dict_a, dict_b):
    """
    Add dict_a into dict_b
    Merge values if possible else dict_a value replace dict_b value
    """
    for key in dict_a:
        if key in dict_b:
            if isinstance(dict_a[key], dict) and isinstance(dict_b[key], dict):
                merge_dict(dict_a[key], dict_b[key])
            else:
                dict_b[key] = dict_a[key]
        else:
            dict_b[key] = dict_a[key]
    return dict_b


def merge_zuul_list(dict_a, dict_b, key):
    value_a = dict_a.get('zuul', {}).get(key, [])
    value_b = dict_b.get('zuul', {}).get(key, [])
    if not isinstance(value_a, list):
        value_a = []
    if not isinstance(value_b, list):
        value_b = []
    return value_a + value_b


def merge_data(dict_a, dict_b):
    """
    Merge dict_a into dict_b, handling any special cases for zuul variables
    """
    artifacts = merge_zuul_list(dict_a, dict_b, 'artifacts')
    file_comments = merge_file_comments(dict_a, dict_b)
    warnings = merge_zuul_list(dict_a, dict_b, 'warnings')
    retry = dict_a.get('zuul', {}).get('retry')
    merge_dict(dict_a, dict_b)
    if artifacts:
        dict_b.setdefault('zuul', {})['artifacts'] = artifacts
    if file_comments:
        dict_b.setdefault("zuul", {})["file_comments"] = file_comments
    if warnings:
        dict_b.setdefault('zuul', {})['warnings'] = warnings
    if retry:
        dict_b.setdefault('zuul', {})['retry'] = retry

    return dict_b


def merge_file_comments(dict_a, dict_b):
    """Merge file_comments from both dictionary.

    File comments applied to the same line and/or file will be added to the
    existing ones.
    """
    file_comments_a = dict_a.get('zuul', {}).get("file_comments", {})
    if not isinstance(file_comments_a, dict):
        file_comments_a = {}
    file_comments_b = dict_b.get('zuul', {}).get("file_comments", {})
    if not isinstance(file_comments_b, dict):
        file_comments_b = {}
    # Merge all comments for each file from b into a. In case a contains
    # comments for the same file, both are merged and not overriden.
    file_comments = deepcopy(file_comments_b)
    for key, value in file_comments_a.items():
        file_comments.setdefault(key, []).extend(value)
    return file_comments


def set_value(path, new_data, new_file, new_secret_data, new_secret_file):
    workdir = os.path.dirname(path)
    data = None
    secret_data = None

    # Read any existing zuul_return data.
    if os.path.exists(path):
        with open(path, 'r') as f:
            file_data = f.read()
    if file_data:
        file_data = json.loads(file_data)
        data = file_data['data']
        secret_data = file_data['secret_data']
    else:
        data = {}
        secret_data = {}

    # If a file of data was supplied, merge its contents.
    if new_file:
        with open(new_file, 'r') as f:
            merge_data(json.load(f), data)
    if new_secret_file:
        with open(new_secret_file, 'r') as f:
            merge_data(json.load(f), secret_data)

    # If a 'data' value was supplied, merge it.
    if new_data:
        merge_data(new_data, data)
    if new_secret_data:
        merge_data(new_secret_data, secret_data)

    # Validate the schema:
    validate_schema(data)
    validate_schema(secret_data)

    # Replace our results file ('path') with the updated data.
    (f, tmp_path) = tempfile.mkstemp(dir=workdir)
    try:
        f = os.fdopen(f, 'w')
        json.dump({'data': data, 'secret_data': secret_data}, f)
        f.close()
        os.rename(tmp_path, path)
    except Exception:
        os.unlink(tmp_path)
        raise


class ActionModule(ActionBase):
    def run(self, tmp=None, task_vars=None):
        """
        Implementation of our action plugin.

        Our plugin currently accepts these arguments:

           data - A dictionary of arbitrary data to return to Zuul.
           secret_data - The same, but secret data.
           path - File location on the executor to store the return data.
                  Unlikely to be supplied.
           file - A JSON-formatted file storing the data to return to Zuul.
                  This can be used instead of, or in conjunction with, the
                  'data' argument to return large amounts of data.
           secret_file - The same, but secret data.

        Note: The plugin parameters are stored in the self._task.args variable.

        :param tmp: Deprecated parameter.
        :param task_vars: The variables (host vars, group vars, config vars,
            etc) associated with this task.

        :returns: Dictionary of results from the plugin.
        """
        if task_vars is None:
            task_vars = dict()
        results = super(ActionModule, self).run(tmp, task_vars)
        del tmp  # tmp no longer has any effect

        path = self._task.args.get('path')
        if not path:
            path = os.path.join(os.environ['ZUUL_JOBDIR'], 'work',
                                'results.json')

        set_value(
            path,
            self._task.args.get('data'),
            self._task.args.get('file'),
            self._task.args.get('secret_data'),
            self._task.args.get('secret_file'))

        return results



# Copyright 2014 OpenStack Foundation
# Copyright 2014 Hewlett-Packard Development Company, L.P.
# Copyright 2016 Red Hat
# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
import os
import socket
import threading

from zuul.lib.queue import NamedQueue


class Command:
    name = None
    help = None
    args = []


class Argument:
    name = None
    help = None
    required = None
    default = None


class StopCommand(Command):
    name = 'stop'
    help = 'Stop the running process'


class GracefulCommand(Command):
    name = 'graceful'
    help = 'Stop after completing existing work'


class PauseCommand(Command):
    name = 'pause'
    help = 'Stop accepting new work'


class UnPauseCommand(Command):
    name = 'unpause'
    help = 'Resume accepting new work'


class ReplCommand(Command):
    name = 'repl'
    help = 'Enable the REPL for debugging'


class NoReplCommand(Command):
    name = 'norepl'
    help = 'Disable the REPL'


class CommandSocket(object):
    log = logging.getLogger("zuul.CommandSocket")

    def __init__(self, path):
        self.running = False
        self.path = path
        self.queue = NamedQueue('CommandSocketQueue')

    def start(self):
        self.running = True
        if os.path.exists(self.path):
            os.unlink(self.path)
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.socket.bind(self.path)
        self.socket.listen(1)
        self.socket_thread = threading.Thread(target=self._socketListener)
        self.socket_thread.daemon = True
        self.socket_thread.start()

    def stop(self):
        # First, wake up our listener thread with a connection and
        # tell it to stop running.
        self.running = False
        with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as s:
            s.connect(self.path)
            s.sendall(b'_stop\n')
        # The command '_stop' will be ignored by our listener, so
        # directly inject it into the queue so that consumers of this
        # class which are waiting in .get() are awakened.  They can
        # either handle '_stop' or just ignore the unknown command and
        # then check to see if they should continue to run before
        # re-entering their loop.
        self.queue.put(('_stop', []))
        self.socket_thread.join()

    def _socketListener(self):
        while self.running:
            try:
                s, addr = self.socket.accept()
                self.log.debug("Accepted socket connection %s" % (s,))
                buf = b''
                while True:
                    buf += s.recv(1)
                    if buf[-1:] == b'\n':
                        break
                buf = buf.strip()
                self.log.debug("Received %s from socket" % (buf,))
                s.close()

                buf = buf.decode('utf8')
                parts = buf.split(' ', 1)
                # Because we use '_stop' internally to wake up a
                # waiting thread, don't allow it to actually be
                # injected externally.
                args = parts[1:]
                if args:
                    args = json.loads(args[0])
                if parts[0] != '_stop':
                    self.queue.put((parts[0], args))
            except Exception:
                self.log.exception("Exception in socket handler")

        # Unlink socket file within the thread so join works and we don't
        # leak the socket file.
        self.socket.close()
        if os.path.exists(self.path):
            os.unlink(self.path)

    def get(self):
        if not self.running:
            raise Exception("CommandSocket.get called while stopped")
        return self.queue.get()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""buildset_event_table

Revision ID: 6c1582c1d08c
Revises: ac1dad8c9434
Create Date: 2024-03-25 12:28:58.885794

"""

# revision identifiers, used by Alembic.
revision = '6c1582c1d08c'
down_revision = 'ac1dad8c9434'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

BUILDSET_EVENT_TABLE = "zuul_buildset_event"
BUILDSET_TABLE = "zuul_buildset"


def upgrade(table_prefix=''):
    prefixed_buildset_event = table_prefix + BUILDSET_EVENT_TABLE
    prefixed_buildset = table_prefix + BUILDSET_TABLE

    op.create_table(
        prefixed_buildset_event,
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("buildset_id", sa.Integer,
                  sa.ForeignKey(
                      f'{prefixed_buildset}.id',
                      name=f'{prefixed_buildset_event}_buildset_id_fkey',
                  )),
        sa.Column("event_time", sa.DateTime),
        sa.Column("event_type", sa.String(255)),
        sa.Column("description", sa.TEXT()),
    )
    op.create_index(
        f'{prefixed_buildset_event}_buildset_id_idx',
        prefixed_buildset_event,
        ['buildset_id'])


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2017 Red Hat, Inc.
# Copyright 2021-2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import logging
import socket
import ssl
import threading
from configparser import ConfigParser
from typing import Optional

from zuul.exceptions import StreamingError
from zuul.lib import streamer_utils
from zuul.lib import commandsocket
from zuul.lib.config import get_default
from zuul.lib.monitoring import MonitoringServer
from zuul.version import get_version_string
from zuul.zk import ZooKeeperClient
from zuul.zk.components import COMPONENT_REGISTRY, FingerGatewayComponent
from zuul.zk.executor import ExecutorApi

COMMANDS = [
    commandsocket.StopCommand,
]


class RequestHandler(streamer_utils.BaseFingerRequestHandler):
    '''
    Class implementing the logic for handling a single finger request.
    '''

    log = logging.getLogger("zuul.fingergw")

    def __init__(self, *args, **kwargs):
        self.fingergw = kwargs.pop('fingergw')
        super(RequestHandler, self).__init__(*args, **kwargs)

    def _readSocket(self, sock, build_uuid):
        # timeout only on the connection, let recv() wait forever
        sock.settimeout(None)
        msg = "%s\n" % build_uuid    # Must have a trailing newline!
        sock.sendall(msg.encode('utf-8'))
        while True:
            data = sock.recv(1024)
            if data:
                self.request.sendall(data)
            else:
                break

    def _fingerClient(self, server, port, build_uuid, use_ssl):
        '''
        Open a finger connection and return all streaming results.

        :param server: The remote server.
        :param port: The remote port.
        :param build_uuid: The build UUID to stream.

        Both IPv4 and IPv6 are supported.
        '''
        with socket.create_connection((server, port), timeout=10) as s:
            if use_ssl:
                context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                context.verify_mode = ssl.CERT_REQUIRED
                context.check_hostname = self.fingergw.tls_verify_hostnames
                context.load_cert_chain(self.fingergw.tls_cert,
                                        self.fingergw.tls_key)
                context.load_verify_locations(self.fingergw.tls_ca)
                with context.wrap_socket(s, server_hostname=server) as s:
                    self._readSocket(s, build_uuid)
            else:
                self._readSocket(s, build_uuid)

    def handle(self):
        '''
        This method is called by the socketserver framework to handle an
        incoming request.
        '''
        server = None
        port = None
        try:
            build_uuid = self.getCommand()
            port_location = streamer_utils.getJobLogStreamAddress(
                self.fingergw.executor_api,
                build_uuid, source_zone=self.fingergw.zone)

            if not port_location:
                msg = 'Invalid build UUID %s' % build_uuid
                self.request.sendall(msg.encode('utf-8'))
                return

            server = port_location['server']
            port = port_location['port']
            use_ssl = port_location.get('use_ssl', False)
            self._fingerClient(server, port, build_uuid, use_ssl)
        except StreamingError as e:
            self.request.sendall(str(e).encode("utf-8"))
        except BrokenPipeError:   # Client disconnect
            return
        except Exception:
            self.log.exception(
                'Finger request handling exception (%s:%s):',
                server, port)
            msg = 'Internal streaming error'
            self.request.sendall(msg.encode('utf-8'))
            return


class FingerGateway(object):
    '''
    Class implementing the finger multiplexing/gateway logic.

    For each incoming finger request, a new thread is started that will
    be responsible for finding which Zuul executor is executing the
    requested build (by asking ZooKeeper), forwarding the request to that
    executor, and streaming the results back to our client.
    '''

    log = logging.getLogger("zuul.fingergw")
    handler_class = RequestHandler

    def __init__(
        self,
        config: ConfigParser,
        command_socket: Optional[str],
        pid_file: Optional[str],
    ):
        '''
        Initialize the finger gateway.

        :param config: The parsed Zuul configuration.
        :param str command_socket: Path to the daemon command socket.
        :param str pid_file: Path to the daemon PID file.
        '''

        host = get_default(config, 'fingergw', 'listen_address', '::')
        self.port = int(get_default(config, 'fingergw', 'port', 79))
        self.public_port = int(get_default(
            config, 'fingergw', 'public_port', self.port))
        user = get_default(config, 'fingergw', 'user', None)

        self.address = (host, self.port)
        self.user = user
        self.pid_file = pid_file

        self.server = None
        self.server_thread = None

        self.command_thread = None
        self.command_running = False
        self.command_socket_path = command_socket
        self.command_socket = None

        self.tls_key = get_default(config, 'fingergw', 'tls_key')
        self.tls_cert = get_default(config, 'fingergw', 'tls_cert')
        self.tls_ca = get_default(config, 'fingergw', 'tls_ca')
        self.tls_verify_hostnames = get_default(
            config, 'fingergw', 'tls_verify_hostnames', default=True)
        client_only = get_default(config, 'fingergw', 'tls_client_only',
                                  default=False)
        if (all([self.tls_key, self.tls_cert, self.tls_ca])
            and not client_only):
            self.tls_listen = True
        else:
            self.tls_listen = False

        self.command_map = {
            commandsocket.StopCommand.name: self.stop,
        }

        self.hostname = get_default(config, 'fingergw', 'hostname',
                                    socket.getfqdn())
        self.zone = get_default(config, 'fingergw', 'zone')

        self.zk_client = ZooKeeperClient.fromConfig(config)
        self.zk_client.connect()
        self.component_info = FingerGatewayComponent(
            self.zk_client, self.hostname, version=get_version_string()
        )
        if self.zone is not None:
            self.component_info.zone = self.zone
            self.component_info.public_port = self.public_port
        if self.tls_listen:
            self.component_info.use_ssl = True
        self.component_info.register()
        COMPONENT_REGISTRY.create(self.zk_client)

        self.monitoring_server = MonitoringServer(config, 'fingergw',
                                                  self.component_info)
        self.monitoring_server.start()

        self.executor_api = ExecutorApi(self.zk_client, use_cache=False)

    def _runCommand(self):
        while self.command_running:
            try:
                command, args = self.command_socket.get()
                if command != '_stop':
                    self.command_map[command](*args)
                else:
                    return
            except Exception:
                self.log.exception("Exception while processing command")

    def _run(self):
        try:
            self.server.serve_forever()
        except Exception:
            self.log.exception('Abnormal termination:')
            raise

    def start(self):
        self.log.info("Starting finger gateway")
        kwargs = dict(
            user=self.user,
            pid_file=self.pid_file,
        )
        if self.tls_listen:
            kwargs.update(dict(
                server_ssl_ca=self.tls_ca,
                server_ssl_cert=self.tls_cert,
                server_ssl_key=self.tls_key,
            ))

        self.server = streamer_utils.CustomThreadingTCPServer(
            self.address,
            functools.partial(self.handler_class, fingergw=self),
            **kwargs)

        # Update port that we really use if we configured a port of 0
        if self.public_port == 0:
            self.public_port = self.server.socket.getsockname()[1]
            self.component_info.public_port = self.public_port

        # Start the command processor after the server and privilege drop
        if self.command_socket_path:
            self.log.debug("Starting command processor")
            self.command_socket = commandsocket.CommandSocket(
                self.command_socket_path)
            self.command_socket.start()
            self.command_running = True
            self.command_thread = threading.Thread(
                target=self._runCommand, name='command')
            self.command_thread.daemon = True
            self.command_thread.start()

        # The socketserver shutdown() call will hang unless the call
        # to server_forever() happens in another thread. So let's do that.
        self.server_thread = threading.Thread(target=self._run)
        self.server_thread.daemon = True
        self.server_thread.start()
        self.component_info.state = self.component_info.RUNNING

        self.log.info("Finger gateway is started")

    def stop(self):
        self.log.info("Stopping finger gateway")
        self.component_info.state = self.component_info.STOPPED

        if self.server:
            try:
                self.server.shutdown()
                self.server.server_close()
                self.server = None
            except Exception:
                self.log.exception("Error stopping TCP server:")

        if self.command_socket:
            self.command_running = False

            try:
                self.command_socket.stop()
            except Exception:
                self.log.exception("Error stopping command socket:")

        self.zk_client.disconnect()
        self.monitoring_server.stop()

        self.log.info("Finger gateway is stopped")

    def join(self):
        '''
        Wait on the gateway to shutdown.
        '''
        self.server_thread.join()

        if self.command_thread:
            self.command_thread.join()
        self.monitoring_server.join()



# This file is derived from the Kazoo project
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import namedtuple


class AddWatchMode(object):
    """Modes for use with :meth:`~kazoo.client.KazooClient.add_watch`

    .. attribute:: PERSISTENT

        The watch is not removed when trigged.

    .. attribute:: PERSISTENT_RECURSIVE

        The watch is not removed when trigged, and applies to all
        paths underneath the supplied path as well.
    """

    PERSISTENT = 0
    PERSISTENT_RECURSIVE = 1


class WatcherType(object):
    """Watcher types for use with
    :meth:`~kazoo.client.KazooClient.remove_all_watches`

    .. attribute:: CHILDREN

        Child watches.

    .. attribute:: DATA

        Data watches.

    .. attribute:: ANY

        Any type of watch (child, data, persistent, or persistent
        recursive).

    """

    CHILDREN = 1
    DATA = 2
    ANY = 3


class WatchedEvent(namedtuple("WatchedEvent",
                              ("type", "state", "path", "zxid"))):
    """A change on ZooKeeper that a Watcher is able to respond to.

    The :class:`WatchedEvent` includes exactly what happened, the
    current state of ZooKeeper, and the path of the node that was
    involved in the event. An instance of :class:`WatchedEvent` will be
    passed to registered watch functions.

    .. attribute:: type

        A :class:`EventType` attribute indicating the event type.

    .. attribute:: state

        A :class:`KeeperState` attribute indicating the Zookeeper
        state.

    .. attribute:: path

        The path of the node for the watch event.

    .. attribute:: zxid

        The zxid of the transaction that triggered this watch if it is
        of one of the following types:

        * EventType.CREATED
        * EventType.DELETED
        * EventType.CHANGED
        * EventType.CHILD

        Otherwise, returns WatchedEvent.NO_ZXID. Note that NO_ZXID is also
        returned by old servers that do not support this feature.

    """

    NO_ZXID = -1



# Copyright 2018-2019 Red Hat, Inc.
# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import voluptuous as vs


artifact = {
    vs.Required('name'): str,
    vs.Required('url'): str,
    'metadata': dict,
}

artifact_data = {
    'zuul': {
        'log_url': str,
        'artifacts': [artifact],
        vs.Extra: object,
    },
    vs.Extra: object,
}

warning_data = {
    'zuul': {
        'log_url': str,
        'warnings': [str],
        vs.Extra: object,
    },
    vs.Extra: object,
}

artifact_schema = vs.Schema(artifact_data)
warning_schema = vs.Schema(warning_data)



# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import socket
import threading

import prometheus_client

from zuul.lib.config import get_default


def _get_best_family(address, port):
    """Automatically select address family depending on address"""
    # HTTPServer defaults to AF_INET, which will not start properly if
    # binding an ipv6 address is requested.
    # This function is based on what upstream python did for http.server
    # in https://github.com/python/cpython/pull/11767
    infos = socket.getaddrinfo(address, port)
    family, _, _, _, sockaddr = next(iter(infos))
    return family, sockaddr[0]


class MonitoringServer:
    def __init__(self, config, section, component_info):
        if not config.has_option(section, 'prometheus_port'):
            self.httpd = None
            return

        self.component_info = component_info
        port = int(config.get(section, 'prometheus_port'))
        addr = get_default(
            config, section, 'prometheus_addr', '0.0.0.0')

        self.prometheus_app = prometheus_client.make_wsgi_app(
            prometheus_client.registry.REGISTRY)

        class TmpServer(prometheus_client.exposition.ThreadingWSGIServer):
            """Copy of ThreadingWSGIServer to update address_family locally"""
        TmpServer.address_family, addr = _get_best_family(addr, port)

        self.httpd = prometheus_client.exposition.make_server(
            addr, port,
            self.handleRequest,
            TmpServer,
            handler_class=prometheus_client.exposition._SilentHandler)
        # The unit tests pass in 0 for the port
        self.port = self.httpd.socket.getsockname()[1]

    def start(self):
        if self.httpd is None:
            return
        self.thread = threading.Thread(target=self.httpd.serve_forever)
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        if self.httpd is None:
            return
        self.httpd.shutdown()

    def join(self):
        if self.httpd is None:
            return
        self.thread.join()
        self.httpd.socket.close()

    def handleRequest(self, environ, start_response):
        headers = []
        output = b''
        if environ['PATH_INFO'] == '/health/live':
            status = '200 OK'
        elif environ['PATH_INFO'] == '/health/ready':
            if (self.component_info.state in (
                    self.component_info.RUNNING,
                    self.component_info.PAUSED)):
                status = '200 OK'
            else:
                status = '503 Service Unavailable'
        elif environ['PATH_INFO'] == '/health/status':
            status = '200 OK'
            headers = [('Content-Type', 'text/plain')]
            output = str(self.component_info.state).encode('utf8').upper()
        elif environ['PATH_INFO'] in ('/metrics', '/'):
            # The docs say '/metrics' but '/' worked and was likely
            # used by users, so let's support both for now.
            return self.prometheus_app(environ, start_response)
        else:
            status = '404 Not Found'
        # Return output
        start_response(status, headers)
        return [output]



# Copyright 2018 EasyStack, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import configparser

from tests.base import BaseTestCase
from tests.base import FIXTURE_DIR
from zuul.lib.config import get_default


class TestDefaultConfigValue(BaseTestCase):
    config_file = 'zuul.conf'

    def setUp(self):
        super(TestDefaultConfigValue, self).setUp()
        self.config = configparser.ConfigParser()
        self.config.read(os.path.join(FIXTURE_DIR, self.config_file))

    def test_default_config_value(self):
        default_value = get_default(self.config,
                                    'web',
                                    'static_cache_expiry',
                                    default=3600)
        self.assertEqual(1200, default_value)



# Copyright 2021 BMW Group
# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
import time
from contextlib import suppress
from enum import Enum
import zlib

from kazoo.exceptions import LockTimeout, NoNodeError, NodeExistsError
from kazoo.protocol.states import EventType, ZnodeStat
from kazoo.client import TransactionRequest

from zuul.lib.jsonutil import json_dumps
from zuul.lib.logutil import get_annotated_logger
from zuul.model import JobRequest
from zuul.zk import sharding
from zuul.zk.event_queues import JobResultFuture
from zuul.zk.exceptions import JobRequestNotFound
from zuul.zk.locks import SessionAwareLock
from zuul.zk.cache import ZuulTreeCache


class JobRequestEvent(Enum):
    CREATED = 0
    UPDATED = 1
    RESUMED = 2
    CANCELED = 3
    DELETED = 4


class RequestUpdater:
    """This class cooperates with the event queues so that we can update a
    request and submit an event in a single transaction."""

    _log = logging.getLogger("zuul.JobRequestQueue")

    def __init__(self, request):
        self.request = request
        self.log = get_annotated_logger(
            self._log, event=request.event_id, build=request.uuid
        )

    def preRun(self):
        """A pre-flight check.  Return whether we should attempt the
        transaction."""
        self.log.debug("Updating request %s", self.request)

        if self.request._zstat is None:
            self.log.debug(
                "Cannot update request %s: Missing version information.",
                self.request.uuid,
            )
            return False
        return True

    def run(self, client):
        """Actually perform the transaction.  The 'client' argument may be a
        transaction or a plain client."""
        if isinstance(client, TransactionRequest):
            setter = client.set_data
        else:
            setter = client.set
        return setter(
            self.request.path,
            JobRequestQueue._dictToBytes(self.request.toDict()),
            version=self.request._zstat.version,
        )

    def postRun(self, result):
        """Process the results of the transaction."""
        try:
            if isinstance(result, Exception):
                raise result
            elif isinstance(result, ZnodeStat):
                self.request._zstat = result
            else:
                raise Exception("Unknown result from ZooKeeper for %s: %s",
                                self.request, result)
        except NoNodeError:
            raise JobRequestNotFound(
                f"Could not update {self.request.path}"
            )


class JobRequestCache(ZuulTreeCache):
    def __init__(self, client, root, request_class,
                 request_callback=None,
                 event_callback=None):
        self.request_class = request_class
        self.request_callback = request_callback
        self.event_callback = event_callback
        super().__init__(client, root)

    def _parsePath(self, path):
        if not path.startswith(self.root):
            return None
        path = path[len(self.root) + 1:]
        parts = path.split('/')
        # We are interested in requests with a parts that look like:
        # ([<self.items_path>, <self.locks_path>], <uuid>, ...)
        if len(parts) < 2:
            return None
        return parts

    def parsePath(self, path):
        key = None
        fetch = False
        parts = self._parsePath(path)
        if parts is None:
            return (key, fetch)
        if len(parts) != 2:
            return (key, fetch)
        if parts[0] == 'requests':
            key = (parts[0], parts[1])
            fetch = True
        return (key, fetch)

    def preCacheHook(self, event, exists, data=None, stat=None):
        parts = self._parsePath(event.path)
        if parts is None:
            return

        if parts[0] == 'requests' and self.event_callback:
            if event.type == EventType.CREATED:
                if len(parts) == 3:
                    job_event = None
                    if parts[2] == 'cancel':
                        job_event = JobRequestEvent.CANCELED
                    elif parts[2] == 'resume':
                        job_event = JobRequestEvent.RESUMED
                    if job_event:
                        request = self.getRequest(parts[1])
                        if not request:
                            return
                        self.event_callback(request, job_event)
            elif event.type == EventType.DELETED:
                if len(parts) == 2:
                    request = self.getRequest(parts[1])
                    if not request:
                        return
                    self.event_callback(request, JobRequestEvent.DELETED)
        elif parts[0] == 'locks':
            request = self.getRequest(parts[1])
            if not request:
                return
            if len(parts) < 3:
                return
            if exists:
                request.lock_contenders += 1
            else:
                request.lock_contenders -= 1
            request.is_locked = bool(request.lock_contenders)

    def objectFromRaw(self, key, data, stat):
        if key[0] == 'requests':
            content = self._bytesToDict(data)
            request = self.request_class.fromDict(content)
            request._zstat = stat
            request.path = "/".join([self.root, 'requests', request.uuid])
            request._old_state = request.state
            return request

    def updateFromRaw(self, request, key, data, stat):
        content = self._bytesToDict(data)
        with request.thread_lock:
            # TODO: move thread locking into the TreeCache so we don't
            # duplicate this check.
            if stat.mzxid >= request._zstat.mzxid:
                request.updateFromDict(content)
                request._zstat = stat

    def postCacheHook(self, event, data, stat, key, obj):
        if key[0] == 'requests' and self.request_callback:
            if event.type in (EventType.CREATED, EventType.NONE):
                self.request_callback()

            # This is a test-specific condition: For test cases which
            # are using hold_*_jobs_in_queue the state change on the
            # request from HOLD to REQUESTED is done outside of the
            # server.  Thus, we must also set the wake event (the
            # callback) so the servercan pick up those jobs after they
            # are released. To not cause a thundering herd problem in
            # production for each cache update, the callback is only
            # called under this very specific condition that can only
            # occur in the tests.
            elif (
                self.request_callback
                and obj
                and obj._old_state == self.request_class.HOLD
                and obj.state == self.request_class.REQUESTED
            ):
                obj._old_state = obj.state
                self.request_callback()

    def getRequest(self, request_uuid):
        key = ('requests', request_uuid)
        return self._cached_objects.get(key)

    def getRequests(self):
        return self._cached_objects.values()


class JobRequestQueue:
    log = logging.getLogger("zuul.JobRequestQueue")
    request_class = JobRequest

    def __init__(self, client, root, use_cache=True,
                 request_callback=None, event_callback=None):
        self.zk_client = client
        self.kazoo_client = client.client

        self.REQUEST_ROOT = f"{root}/requests"
        self.LOCK_ROOT = f"{root}/locks"
        self.PARAM_ROOT = f"{root}/params"
        self.RESULT_ROOT = f"{root}/results"
        self.RESULT_DATA_ROOT = f"{root}/result-data"
        self.WAITER_ROOT = f"{root}/waiters"

        self.kazoo_client.ensure_path(self.REQUEST_ROOT)
        self.kazoo_client.ensure_path(self.PARAM_ROOT)
        self.kazoo_client.ensure_path(self.RESULT_ROOT)
        self.kazoo_client.ensure_path(self.RESULT_DATA_ROOT)
        self.kazoo_client.ensure_path(self.WAITER_ROOT)
        self.kazoo_client.ensure_path(self.LOCK_ROOT)

        if use_cache:
            self.cache = JobRequestCache(
                client, root, self.request_class,
                request_callback, event_callback)
        else:
            self.cache = None

    # So far only used by tests
    def waitForSync(self):
        self.cache.waitForSync()

    @property
    def initial_state(self):
        # This supports holding requests in tests
        return self.request_class.REQUESTED

    def inState(self, *states):
        if not states:
            # If no states are provided, build a tuple containing all available
            # ones to always match. We need a tuple to be compliant to the
            # type of *states above.
            states = self.request_class.ALL_STATES

        requests = [
            req for req in list(self.cache.getRequests())
            if req.state in states
        ]

        # Sort the list of requests by precedence and their creation time
        # in ZooKeeper in ascending order to prevent older requests from
        # starving.
        return sorted(requests)

    def next(self):
        for request in self.inState(self.request_class.REQUESTED):
            request = self.cache.getRequest(request.uuid)
            if (request and
                request.state == self.request_class.REQUESTED and
                not request.is_locked):
                yield request

    def submit(self, request, params, needs_result=False):
        log = get_annotated_logger(self.log, event=request.event_id)

        path = "/".join([self.REQUEST_ROOT, request.uuid])
        request.path = path

        if not isinstance(request, self.request_class):
            raise RuntimeError("Request of wrong class")
        if request.state != self.request_class.UNSUBMITTED:
            raise RuntimeError("Request state must be unsubmitted")
        request.state = self.initial_state

        result = None

        # If a result is needed, create the result_path with the same
        # UUID and store it on the request, so the server can store
        # the result there.
        if needs_result:
            result_path = "/".join(
                [self.RESULT_ROOT, request.uuid]
            )
            waiter_path = "/".join(
                [self.WAITER_ROOT, request.uuid]
            )
            self.kazoo_client.create(waiter_path, ephemeral=True)
            result = JobResultFuture(self.zk_client, request.path,
                                     result_path, waiter_path)
            request.result_path = result_path

        log.debug("Submitting job request to ZooKeeper %s", request)

        params_path = self._getParamsPath(request.uuid)
        with sharding.BufferedShardWriter(
            self.kazoo_client, params_path
        ) as stream:
            stream.write(zlib.compress(self._dictToBytes(params)))

        self.kazoo_client.create(path, self._dictToBytes(request.toDict()))

        return result

    def getRequestUpdater(self, request):
        return RequestUpdater(request)

    def update(self, request):
        updater = self.getRequestUpdater(request)
        if not updater.preRun():
            return

        try:
            result = updater.run(self.kazoo_client)
        except Exception as e:
            result = e

        updater.postRun(result)

    def reportResult(self, request, result):
        # Write the result data first since it may be multiple nodes.
        result_data_path = "/".join(
            [self.RESULT_DATA_ROOT, request.uuid]
        )
        with sharding.BufferedShardWriter(
                self.kazoo_client, result_data_path) as stream:
            stream.write(zlib.compress(self._dictToBytes(result)))

        # Then write the result node to signify it's ready.
        data = {'result_data_path': result_data_path}
        self.kazoo_client.create(request.result_path,
                                 self._dictToBytes(data))

    def getRequest(self, request_uuid):
        if self.cache:
            return self.cache.getRequest(request_uuid)
        else:
            path = "/".join([self.REQUEST_ROOT, request_uuid])
            return self._get(path)

    def _get(self, path):
        try:
            data, zstat = self.kazoo_client.get(path)
        except NoNodeError:
            return None

        if not data:
            return None

        content = self._bytesToDict(data)
        request = self.request_class.fromDict(content)
        request.path = path
        request._zstat = zstat
        return request

    def refresh(self, request):
        """Refreshs a request object with the current data from ZooKeeper. """
        try:
            data, zstat = self.kazoo_client.get(request.path)
        except NoNodeError:
            raise JobRequestNotFound(
                f"Could not refresh {request}, ZooKeeper node is missing")

        if not data:
            raise JobRequestNotFound(
                f"Could not refresh {request}, ZooKeeper node is empty")

        content = self._bytesToDict(data)

        with request.thread_lock:
            if zstat.mzxid >= request._zstat.mzxid:
                request.updateFromDict(content)
                request._zstat = zstat

    def remove(self, request):
        log = get_annotated_logger(self.log, request.event_id)
        log.debug("Removing request %s", request)
        try:
            self.kazoo_client.delete(request.path, recursive=True)
        except NoNodeError:
            # Nothing to do if the node is already deleted
            pass
        try:
            self.clearParams(request)
        except NoNodeError:
            pass
        self._deleteLock(request.uuid)

    # We use child nodes here so that we don't need to lock the
    # request node.
    def requestResume(self, request):
        try:
            self.kazoo_client.create(f"{request.path}/resume")
        except (NoNodeError, NodeExistsError):
            # znode is either gone or already exists
            pass

    def requestCancel(self, request):
        try:
            self.kazoo_client.create(f"{request.path}/cancel")
        except (NoNodeError, NodeExistsError):
            # znode is either gone or already exists
            pass

    def fulfillResume(self, request):
        self.kazoo_client.delete(f"{request.path}/resume")

    def fulfillCancel(self, request):
        self.kazoo_client.delete(f"{request.path}/cancel")

    def lock(self, request, blocking=True, timeout=None):
        path = "/".join([self.LOCK_ROOT, request.uuid])
        have_lock = False
        lock = None
        try:
            lock = SessionAwareLock(self.kazoo_client, path)
            have_lock = lock.acquire(blocking, timeout)
        except NoNodeError:
            # Request disappeared
            have_lock = False
        except LockTimeout:
            have_lock = False
            self.log.error(
                "Timeout trying to acquire lock: %s", request.uuid
            )

        # If we aren't blocking, it's possible we didn't get the lock
        # because someone else has it.
        if not have_lock:
            return False

        if not self.kazoo_client.exists(request.path):
            self._releaseLock(request, lock)
            return False

        # Update the request to ensure that we operate on the newest data.
        try:
            self.refresh(request)
        except JobRequestNotFound:
            self._releaseLock(request, lock)
            return False

        request.lock = lock
        return True

    def _releaseLock(self, request, lock):
        """Releases a lock.

        This is used directly after acquiring the lock in case something went
        wrong.
        """
        lock.release()
        self.log.error("Request not found for locking: %s", request.uuid)

        # We may have just re-created the lock parent node just after the
        # scheduler deleted it; therefore we should (re-) delete it.
        self._deleteLock(request.uuid)

    def _deleteLock(self, uuid):
        # Recursively delete the children and the lock parent node.
        path = "/".join([self.LOCK_ROOT, uuid])
        try:
            children = self.kazoo_client.get_children(path)
        except NoNodeError:
            # The lock is apparently already gone.
            return
        tr = self.kazoo_client.transaction()
        for child in children:
            tr.delete("/".join([path, child]))
        tr.delete(path)
        # We don't care about the results
        tr.commit()

    def unlock(self, request):
        if request.lock is None:
            self.log.warning(
                "Request %s does not hold a lock", request
            )
        else:
            request.lock.release()
            request.lock = None

    def isLocked(self, request):
        path = "/".join([self.LOCK_ROOT, request.uuid])
        if not self.kazoo_client.exists(path):
            return False
        lock = SessionAwareLock(self.kazoo_client, path)
        is_locked = len(lock.contenders()) > 0
        return is_locked

    def lostRequests(self, *states):
        # Get a list of requests which are running but not locked by
        # any client.
        if not states:
            states = (self.request_class.RUNNING,)
        for req in self.inState(*states):
            try:
                if req.is_locked:
                    continue
                if self.isLocked(req):
                    continue
                # It may have completed in the interim, so double
                # check that.
                if req.state not in states:
                    continue
                # We may be racing a cache update, so manually
                # refresh and check again.
                self.refresh(req)
                if req.state not in states:
                    continue
            except (NoNodeError, JobRequestNotFound):
                # Request was removed in the meantime
                continue
            yield req

    def _getAllRequestIds(self):
        # Get a list of all request ids without using the cache.
        return self.kazoo_client.get_children(self.REQUEST_ROOT)

    def _findLostParams(self, age):
        # Get data nodes which are older than the specified age (we
        # don't want to delete nodes which are just being written
        # slowly).
        # Convert to MS
        now = int(time.time() * 1000)
        age = age * 1000
        data_nodes = dict()
        for data_id in self.kazoo_client.get_children(self.PARAM_ROOT):
            data_path = self._getParamsPath(data_id)
            data_zstat = self.kazoo_client.exists(data_path)
            if not data_zstat:
                # Node was deleted in the meantime
                continue
            if now - data_zstat.mtime > age:
                data_nodes[data_id] = data_path

        # If there are no candidate data nodes, we don't need to
        # filter them by known requests.
        if not data_nodes:
            return data_nodes.values()

        # Remove current request uuids
        for request_id in self._getAllRequestIds():
            if request_id in data_nodes:
                del data_nodes[request_id]

        # Return the paths
        return data_nodes.values()

    def _findLostResults(self):
        # Get a list of results which don't have a connection waiting for
        # them. As the results and waiters are not part of our cache, we have
        # to look them up directly from ZK.
        waiters1 = set(self.kazoo_client.get_children(self.WAITER_ROOT))
        results = set(self.kazoo_client.get_children(self.RESULT_ROOT))
        result_data = set(self.kazoo_client.get_children(
            self.RESULT_DATA_ROOT))
        waiters2 = set(self.kazoo_client.get_children(self.WAITER_ROOT))

        waiters = waiters1.union(waiters2)
        lost_results = results - waiters
        lost_data = result_data - waiters
        return lost_results, lost_data

    def cleanup(self, age=300):
        # Delete build request params which are not associated with
        # any current build requests.  Note, this does not clean up
        # lost requests themselves; the client takes care of that.
        try:
            for path in self._findLostParams(age):
                try:
                    self.log.error("Removing request params: %s", path)
                    self.kazoo_client.delete(path, recursive=True)
                except Exception:
                    self.log.exception(
                        "Unable to delete request params %s", path)
        except Exception:
            self.log.exception(
                "Error cleaning up request queue %s", self)
        try:
            lost_results, lost_data = self._findLostResults()
            for result_id in lost_results:
                try:
                    path = '/'.join([self.RESULT_ROOT, result_id])
                    self.log.error("Removing request result: %s", path)
                    self.kazoo_client.delete(path, recursive=True)
                except Exception:
                    self.log.exception(
                        "Unable to delete request params %s", result_id)
            for result_id in lost_data:
                try:
                    path = '/'.join([self.RESULT_DATA_ROOT, result_id])
                    self.log.error(
                        "Removing request result data: %s", path)
                    self.kazoo_client.delete(path, recursive=True)
                except Exception:
                    self.log.exception(
                        "Unable to delete request params %s", result_id)
        except Exception:
            self.log.exception(
                "Error cleaning up result queue %s", self)
        try:
            for lock_id in self.kazoo_client.get_children(self.LOCK_ROOT):
                try:
                    lock_path = "/".join([self.LOCK_ROOT, lock_id])
                    request_path = "/".join([self.REQUEST_ROOT, lock_id])
                    if not self.kazoo_client.exists(request_path):
                        self.log.error("Removing stale lock: %s", lock_path)
                        self.kazoo_client.delete(lock_path, recursive=True)
                except Exception:
                    self.log.exception(
                        "Unable to delete lock %s", lock_path)
        except Exception:
            self.log.exception("Error cleaning up locks %s", self)

    @staticmethod
    def _bytesToDict(data):
        return json.loads(data.decode("utf-8"))

    @staticmethod
    def _dictToBytes(data):
        # The custom json_dumps() will also serialize MappingProxyType objects
        return json_dumps(data, sort_keys=True).encode("utf-8")

    def _getParamsPath(self, uuid):
        return '/'.join([self.PARAM_ROOT, uuid])

    def clearParams(self, request):
        """Erase the parameters from ZK to save space"""
        self.kazoo_client.delete(self._getParamsPath(request.uuid),
                                 recursive=True)

    def getParams(self, request):
        """Return the parameters for a request, if they exist.

        Once a request is accepted by an executor, the params
        may be erased from ZK; this will return None in that case.

        """
        with sharding.BufferedShardReader(
            self.kazoo_client, self._getParamsPath(request.uuid)
        ) as stream:
            try:
                data = zlib.decompress(stream.read())
            except NoNodeError:
                return None
            return self._bytesToDict(data)

    def deleteResult(self, path):
        with suppress(NoNodeError):
            self.kazoo_client.delete(path, recursive=True)



# Copyright 2019 Red Hat
# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import os
import git
import yaml
import socket
import time

from zuul.lib import strings
from zuul.zk.layout import LayoutState
from zuul.zk.change_cache import ChangeKey

from tests.base import (
    ZuulTestCase,
    ZuulWebFixture,
    simple_layout,
    skipIfMultiScheduler,
)
from tests.util import random_sha1

from testtools.matchers import MatchesRegex

EMPTY_LAYOUT_STATE = LayoutState("", "", 0, None, {}, -1)


class TestGitlabWebhook(ZuulTestCase):
    config_file = 'zuul-gitlab-driver.conf'

    def setUp(self):
        super().setUp()

        # Start the web server
        self.web = self.useFixture(
            ZuulWebFixture(self.config, self.test_config,
                           self.additional_event_queues, self.upstream_root,
                           self.poller_events,
                           self.git_url_with_auth, self.addCleanup,
                           self.test_root))

        host = '127.0.0.1'
        # Wait until web server is started
        while True:
            port = self.web.port
            try:
                with socket.create_connection((host, port)):
                    break
            except ConnectionRefusedError:
                pass

        self.fake_gitlab.setZuulWebPort(port)

    def tearDown(self):
        super(TestGitlabWebhook, self).tearDown()

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_webhook(self):
        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project', 'master', 'A')
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent(),
                                   use_zuulweb=False,
                                   project='org/project')
        self.waitUntilSettled()

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test1').result)

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_webhook_via_zuulweb(self):
        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project', 'master', 'A')
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent(),
                                   use_zuulweb=True,
                                   project='org/project')
        self.waitUntilSettled()

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test1').result)


class TestGitlabDriver(ZuulTestCase):
    config_file = 'zuul-gitlab-driver.conf'

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_opened(self):

        description = "This is the\nMR description."
        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project', 'master', 'A', description=description)
        self.fake_gitlab.emitEvent(
            A.getMergeRequestOpenedEvent(), project='org/project')
        self.waitUntilSettled()

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test1').result)

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test2').result)

        job = self.getJobFromHistory('project-test2')
        zuulvars = job.parameters['zuul']
        self.assertEqual(str(A.number), zuulvars['change'])
        self.assertEqual(str(A.sha), zuulvars['patchset'])
        self.assertEqual(str(A.sha), zuulvars['commit_id'])
        self.assertEqual('master', zuulvars['branch'])
        self.assertEqual(f'{self.fake_gitlab._test_baseurl}/'
                         'org/project/merge_requests/1',
                         zuulvars['items'][0]['change_url'])
        self.assertEqual(zuulvars["message"], strings.b64encode(description))
        self.assertEqual(2, len(self.history))
        self.assertEqual(2, len(A.notes))
        self.assertEqual(
            A.notes[0]['body'], "Starting check jobs.")
        self.assertThat(
            A.notes[1]['body'],
            MatchesRegex(r'.*project-test1.*SUCCESS.*', re.DOTALL))
        self.assertThat(
            A.notes[1]['body'],
            MatchesRegex(r'.*project-test2.*SUCCESS.*', re.DOTALL))
        self.assertTrue(A.approved)

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_opened_imcomplete(self):

        now = time.monotonic()
        complete_at = now + 3
        with self.fake_gitlab.enable_delayed_complete_mr(complete_at):
            description = "This is the\nMR description."
            A = self.fake_gitlab.openFakeMergeRequest(
                'org/project', 'master', 'A', description=description)
            self.fake_gitlab.emitEvent(
                A.getMergeRequestOpenedEvent(), project='org/project')
            self.waitUntilSettled()

            self.assertEqual('SUCCESS',
                             self.getJobFromHistory('project-test1').result)

            self.assertEqual('SUCCESS',
                             self.getJobFromHistory('project-test2').result)

        self.assertTrue(self.fake_gitlab._test_web_server.stats["get_mr"] > 2)

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_updated(self):

        A = self.fake_gitlab.openFakeMergeRequest('org/project', 'master', 'A')
        mr_tip1_sha = A.sha
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))
        self.assertHistory(
            [
                {'name': 'project-test1', 'changes': '1,%s' % mr_tip1_sha},
                {'name': 'project-test2', 'changes': '1,%s' % mr_tip1_sha},
            ], ordered=False
        )

        self.fake_gitlab.emitEvent(A.getMergeRequestUpdatedEvent())
        mr_tip2_sha = A.sha
        self.waitUntilSettled()
        self.assertEqual(4, len(self.history))
        self.assertHistory(
            [
                {'name': 'project-test1', 'changes': '1,%s' % mr_tip1_sha},
                {'name': 'project-test2', 'changes': '1,%s' % mr_tip1_sha},
                {'name': 'project-test1', 'changes': '1,%s' % mr_tip2_sha},
                {'name': 'project-test2', 'changes': '1,%s' % mr_tip2_sha}
            ], ordered=False
        )

        # Adding or removing reviewer should not trigger a build
        self.fake_gitlab.emitEvent(A.getMergeRequestReviewersUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(4, len(self.history))

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_approved(self):

        A = self.fake_gitlab.openFakeMergeRequest('org/project', 'master', 'A')

        self.fake_gitlab.emitEvent(A.getMergeRequestApprovedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        self.fake_gitlab.emitEvent(A.getMergeRequestUnapprovedEvent())
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

        job = self.getJobFromHistory('project-test-approval')
        zuulvars = job.parameters['zuul']
        self.assertEqual('check-approval', zuulvars['pipeline'])

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_updated_during_build(self):

        A = self.fake_gitlab.openFakeMergeRequest('org/project', 'master', 'A')
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        old = A.sha
        A.addCommit()
        new = A.sha
        self.assertNotEqual(old, new)
        self.waitUntilSettled()

        self.assertEqual(2, len(self.history))
        # MR must not be approved: tested commit isn't current commit
        self.assertFalse(A.approved)

        self.fake_gitlab.emitEvent(A.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()

        self.assertEqual(4, len(self.history))
        self.assertTrue(A.approved)

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_labeled(self):

        A = self.fake_gitlab.openFakeMergeRequest('org/project', 'master', 'A')

        self.fake_gitlab.emitEvent(A.getMergeRequestLabeledEvent(
            add_labels=('label1', 'label2')))
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        self.fake_gitlab.emitEvent(A.getMergeRequestLabeledEvent(
            add_labels=('gateit', )))
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        A.labels = ['verified']
        self.fake_gitlab.emitEvent(A.getMergeRequestLabeledEvent(
            remove_labels=('verified', )))
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_merged(self):

        A = self.fake_gitlab.openFakeMergeRequest('org/project', 'master', 'A')

        self.fake_gitlab.emitEvent(A.getMergeRequestMergedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))
        self.assertHistory([{'name': 'project-promote'}])

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_push_does_not_reconfigure(self):
        # Test that the push event that follows a merge doesn't
        # needlessly trigger reconfiguration.

        A = self.fake_gitlab.openFakeMergeRequest('org/project', 'master', 'A')

        state1 = self.scheds.first.sched.local_layout_state.get("tenant-one")
        self.fake_gitlab.emitEvent(A.getMergeRequestMergedEvent())
        self.fake_gitlab.emitEvent(A.getMergeRequestMergedPushEvent(
            self.fake_gitlab))
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))
        self.assertHistory([{'name': 'project-post-job'},
                            {'name': 'project-promote'}
                            ], ordered=False)
        state2 = self.scheds.first.sched.local_layout_state.get("tenant-one")
        self.assertEqual(state1, state2)

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_push_does_reconfigure(self):
        # Test that the push event that follows a merge does
        # trigger reconfiguration if .zuul.yaml is changed.

        A = self.fake_gitlab.openFakeMergeRequest('org/project', 'master', 'A')

        state1 = self.scheds.first.sched.local_layout_state.get("tenant-one")
        self.fake_gitlab.emitEvent(A.getMergeRequestMergedEvent())
        self.fake_gitlab.emitEvent(A.getMergeRequestMergedPushEvent(
            self.fake_gitlab,
            modified_files=['.zuul.yaml']))
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))
        self.assertHistory([{'name': 'project-post-job'},
                            {'name': 'project-promote'}
                            ], ordered=False)
        state2 = self.scheds.first.sched.local_layout_state.get("tenant-one")
        self.assertNotEqual(state1, state2)

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_updated_builds_aborted(self):

        A = self.fake_gitlab.openFakeMergeRequest('org/project', 'master', 'A')
        mr_tip1_sha = A.sha

        self.executor_server.hold_jobs_in_build = True

        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()

        self.fake_gitlab.emitEvent(A.getMergeRequestUpdatedEvent())
        mr_tip2_sha = A.sha
        self.waitUntilSettled()

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertHistory(
            [
                {'name': 'project-test1', 'result': 'ABORTED',
                 'changes': '1,%s' % mr_tip1_sha},
                {'name': 'project-test2', 'result': 'ABORTED',
                 'changes': '1,%s' % mr_tip1_sha},
                {'name': 'project-test1', 'changes': '1,%s' % mr_tip2_sha},
                {'name': 'project-test2', 'changes': '1,%s' % mr_tip2_sha}
            ], ordered=False
        )

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_dequeue_mr_abandoned(self):
        self.executor_server.hold_jobs_in_build = True

        A = self.fake_gitlab.openFakeMergeRequest('org/project', 'master', 'A')
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.fake_gitlab.emitEvent(A.getMergeRequestClosedEvent())
        self.waitUntilSettled()

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(2, len(self.history))
        self.assertEqual(2, self.countJobResults(self.history, 'ABORTED'))

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_commented(self):

        A = self.fake_gitlab.openFakeMergeRequest('org/project', 'master', 'A')
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

        self.fake_gitlab.emitEvent(
            A.getMergeRequestCommentedEvent('I like that change'))
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

        self.fake_gitlab.emitEvent(
            A.getMergeRequestCommentedEvent('recheck'))
        self.waitUntilSettled()
        self.assertEqual(4, len(self.history))

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    @skipIfMultiScheduler()
    # This test fails reproducibly with multiple schedulers because
    # the fake_gitlab._test_baseurl used in the assertion for the
    # change_url doesn't match.
    # An explanation for this would be that each scheduler (and the
    # test case itself) use different (fake) Gitlab connections.
    # However, the interesting part is that only this test fails
    # although there are other gitlab tests with a similar assertion.
    # Apart from that I'm wondering why this test first fails with
    # multiple schedulers as each scheduler should have a different
    # gitlab connection than the test case itself.
    def test_ref_updated(self):

        event = self.fake_gitlab.getPushEvent('org/project')
        expected_newrev = event[1]['after']
        expected_oldrev = event[1]['before']
        self.fake_gitlab.emitEvent(event)
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))
        self.assertEqual(
            'SUCCESS',
            self.getJobFromHistory('project-post-job').result)

        job = self.getJobFromHistory('project-post-job')
        zuulvars = job.parameters['zuul']
        self.assertEqual('refs/heads/master', zuulvars['ref'])
        self.assertEqual('post', zuulvars['pipeline'])
        self.assertEqual('project-post-job', zuulvars['job'])
        self.assertEqual('master', zuulvars['branch'])
        self.assertEqual(
            f'{self.fake_gitlab._test_baseurl}/org/project/tree/'
            f'{zuulvars["newrev"]}',
            zuulvars['change_url'])
        self.assertEqual(expected_newrev, zuulvars['newrev'])
        self.assertEqual(expected_oldrev, zuulvars['oldrev'])

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_ref_created(self):

        self.create_branch('org/project', 'stable-1.0')
        path = os.path.join(self.upstream_root, 'org/project')
        repo = git.Repo(path)
        newrev = repo.commit('refs/heads/stable-1.0').hexsha
        event = self.fake_gitlab.getPushEvent(
            'org/project', branch='refs/heads/stable-1.0',
            before='0' * 40, after=newrev)
        old = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)
        self.fake_gitlab.emitEvent(event)
        self.waitUntilSettled()
        new = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)
        # New timestamp should be greater than the old timestamp
        self.assertLess(old, new)
        self.assertEqual(1, len(self.history))
        self.assertEqual(
            'SUCCESS',
            self.getJobFromHistory('project-post-job').result)
        job = self.getJobFromHistory('project-post-job')
        zuulvars = job.parameters['zuul']
        self.assertEqual('refs/heads/stable-1.0', zuulvars['ref'])
        self.assertEqual('post', zuulvars['pipeline'])
        self.assertEqual('project-post-job', zuulvars['job'])
        self.assertEqual('stable-1.0', zuulvars['branch'])
        self.assertEqual(newrev, zuulvars['newrev'])

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_ref_deleted(self):

        event = self.fake_gitlab.getPushEvent(
            'org/project', 'stable-1.0', after='0' * 40)
        self.fake_gitlab.emitEvent(event)
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_tag_created(self):

        path = os.path.join(self.upstream_root, 'org/project')
        repo = git.Repo(path)
        repo.create_tag('1.0')
        tagsha = repo.tags['1.0'].commit.hexsha
        event = self.fake_gitlab.getGitTagEvent(
            'org/project', '1.0', tagsha)
        self.fake_gitlab.emitEvent(event)
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))
        self.assertEqual(
            'SUCCESS',
            self.getJobFromHistory('project-tag-job').result)
        job = self.getJobFromHistory('project-tag-job')
        zuulvars = job.parameters['zuul']
        self.assertEqual('refs/tags/1.0', zuulvars['ref'])
        self.assertEqual('tag', zuulvars['pipeline'])
        self.assertEqual('project-tag-job', zuulvars['job'])
        self.assertEqual(tagsha, zuulvars['newrev'])
        self.assertEqual(tagsha, zuulvars['commit_id'])

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_pull_request_with_dyn_reconf(self):
        path = os.path.join(self.upstream_root, 'org/project')
        zuul_yaml = [
            {'job': {
                'name': 'project-test3',
                'run': 'job.yaml'
            }},
            {'project': {
                'check': {
                    'jobs': [
                        'project-test3'
                    ]
                }
            }}
        ]
        playbook = "- hosts: all\n  tasks: []"

        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project', 'master', 'A',
            base_sha=git.Repo(path).head.object.hexsha)
        A.addCommit(
            {'.zuul.yaml': yaml.dump(zuul_yaml),
             'job.yaml': playbook}
        )
        A.addCommit({"dummy.file": ""})
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test1').result)
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test2').result)
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test3').result)

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_pull_request_with_dyn_reconf_alt(self):
        with self.fake_gitlab.enable_uncomplete_mr():
            zuul_yaml = [
                {'job': {
                    'name': 'project-test3',
                    'run': 'job.yaml'
                }},
                {'project': {
                    'check': {
                        'jobs': [
                            'project-test3'
                        ]
                    }
                }}
            ]
            playbook = "- hosts: all\n  tasks: []"
            A = self.fake_gitlab.openFakeMergeRequest(
                'org/project', 'master', 'A')
            A.addCommit(
                {'.zuul.yaml': yaml.dump(zuul_yaml),
                 'job.yaml': playbook}
            )
            A.addCommit({"dummy.file": ""})
            self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
            self.waitUntilSettled()

            self.assertEqual('SUCCESS',
                             self.getJobFromHistory('project-test1').result)
            self.assertEqual('SUCCESS',
                             self.getJobFromHistory('project-test2').result)
            self.assertEqual('SUCCESS',
                             self.getJobFromHistory('project-test3').result)

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_ref_updated_and_tenant_reconfigure(self):

        self.waitUntilSettled()
        old = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)

        zuul_yaml = [
            {'job': {
                'name': 'project-post-job2',
                'run': 'job.yaml'
            }},
            {'project': {
                'post': {
                    'jobs': [
                        'project-post-job2'
                    ]
                }
            }}
        ]
        playbook = "- hosts: all\n  tasks: []"
        self.create_commit(
            'org/project',
            {'.zuul.yaml': yaml.dump(zuul_yaml),
             'job.yaml': playbook},
            message='Add InRepo configuration'
        )
        event = self.fake_gitlab.getPushEvent('org/project',
                                              modified_files=['.zuul.yaml'])
        self.fake_gitlab.emitEvent(event)
        self.waitUntilSettled()

        new = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)
        # New timestamp should be greater than the old timestamp
        self.assertLess(old, new)

        self.assertHistory(
            [{'name': 'project-post-job'},
             {'name': 'project-post-job2'},
            ], ordered=False
        )

    @simple_layout('layouts/crd-gitlab.yaml', driver='gitlab')
    def test_crd_independent(self):

        # Create a change in project1 that a project2 change will depend on
        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project1', 'master', 'A')

        # Create a commit in B that sets the dependency on A
        msg = "Depends-On: %s" % A.url
        B = self.fake_gitlab.openFakeMergeRequest(
            'org/project2', 'master', 'B', description=msg)

        # Make an event to re-use
        self.fake_gitlab.emitEvent(B.getMergeRequestOpenedEvent())
        self.waitUntilSettled()

        # The changes for the job from project2 should include the project1
        # MR content
        changes = self.getJobFromHistory(
            'project2-test', 'org/project2').changes

        self.assertEqual(changes, "%s,%s %s,%s" % (A.number,
                                                   A.sha,
                                                   B.number,
                                                   B.sha))

        # There should be no more changes in the queue
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        self.assertEqual(
            len(tenant.layout.pipeline_managers['check'].state.queues), 0)

    @simple_layout('layouts/requirements-gitlab.yaml', driver='gitlab')
    def test_state_require(self):

        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project1', 'master', 'A')

        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        # Close the MR
        A.closeMergeRequest()

        # A recheck will not trigger the job
        self.fake_gitlab.emitEvent(
            A.getMergeRequestCommentedEvent('recheck'))
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        # Merge the MR
        A.mergeMergeRequest()

        # A recheck will not trigger the job
        self.fake_gitlab.emitEvent(
            A.getMergeRequestCommentedEvent('recheck'))
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        # Re-open the MR
        A.reopenMergeRequest()

        # A recheck will trigger the job
        self.fake_gitlab.emitEvent(
            A.getMergeRequestCommentedEvent('recheck'))
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

    @simple_layout('layouts/requirements-gitlab.yaml', driver='gitlab')
    def test_approval_require(self):

        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project2', 'master', 'A')

        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        A.approved = True

        self.fake_gitlab.emitEvent(A.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        A.approved = False

        self.fake_gitlab.emitEvent(A.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

    @simple_layout('layouts/requirements-gitlab.yaml', driver='gitlab')
    def test_approval_require_community_edition(self):

        with self.fake_gitlab.enable_community_edition():
            A = self.fake_gitlab.openFakeMergeRequest(
                'org/project2', 'master', 'A')

            self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
            self.waitUntilSettled()
            self.assertEqual(0, len(self.history))

            A.approved = True

            self.fake_gitlab.emitEvent(A.getMergeRequestUpdatedEvent())
            self.waitUntilSettled()
            self.assertEqual(1, len(self.history))

            A.approved = False

            self.fake_gitlab.emitEvent(A.getMergeRequestUpdatedEvent())
            self.waitUntilSettled()
            self.assertEqual(1, len(self.history))

    @simple_layout('layouts/requirements-gitlab.yaml', driver='gitlab')
    def test_label_require(self):

        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project3', 'master', 'A')

        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        A.labels = ['gateit', 'prio:low']

        self.fake_gitlab.emitEvent(A.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        A.labels = ['gateit', 'prio:low', 'another_label']

        self.fake_gitlab.emitEvent(A.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

    @simple_layout('layouts/requirements-gitlab.yaml', driver='gitlab')
    def test_state_reject(self):

        fake_mr = self.fake_gitlab.openFakeMergeRequest(project='org/project4',
                                                        branch='master',
                                                        title='Example MR')

        # The job should be triggered for new change
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        # A recheck should not trigger the job for closed change
        fake_mr.closeMergeRequest()
        self.fake_gitlab.emitEvent(
            fake_mr.getMergeRequestCommentedEvent('recheck')
        )
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        # The job should be triggered after change was reopened
        fake_mr.reopenMergeRequest()
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

        # A recheck should not trigger the job after the change was merged
        fake_mr.mergeMergeRequest()
        self.fake_gitlab.emitEvent(
            fake_mr.getMergeRequestCommentedEvent('recheck')
        )
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

    @simple_layout('layouts/requirements-gitlab.yaml', driver='gitlab')
    def test_reject_unapproved(self):

        fake_mr = self.fake_gitlab.openFakeMergeRequest(project='org/project5',
                                                        branch='master',
                                                        title='Example MR')

        # The job should not be triggered for new and unapproved change
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        # The job still should not be triggered for unapproved change
        fake_mr.approved = False
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        # The job now should be triggered for approved change
        fake_mr.approved = True
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        # The job again should not be triggered for unapproved change
        fake_mr.approved = False
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

    @simple_layout('layouts/requirements-gitlab.yaml', driver='gitlab')
    def test_label_reject(self):

        fake_mr = self.fake_gitlab.openFakeMergeRequest(project='org/project6',
                                                        branch='master',
                                                        title='Example MR')

        # The job should not be triggered when required labels are not set
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        # The job still should not be triggered due to missing verified label
        fake_mr.labels = ['gateit', 'ignored']
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        # The job still should not be triggered (now also due to do-not-merge)
        fake_mr.labels = ['gateit', 'ignored', 'do-not-merge']
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        # The job still should not be triggered due to reject condition met
        fake_mr.labels = ['gateit', 'ignored', 'do-not-merge', 'verified']
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        # The job still should not be triggered due to reject condition met
        fake_mr.labels = ['gateit', 'ignored', 'do-not-merge',
                          'do-not-test', 'verified']
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        # The job now should be triggered
        fake_mr.labels = ['gateit', 'ignored', 'verified']
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

    @simple_layout('layouts/requirements-gitlab.yaml', driver='gitlab')
    def test_require_reject_comprehensive(self):

        fake_mr = self.fake_gitlab.openFakeMergeRequest(project='org/project7',
                                                        branch='master',
                                                        title='Example MR')
        fake_mr.approved = False

        # The job should not be triggered for new change without gateit label
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        # The job still should not be triggered even though it is approved now
        fake_mr.approved = True
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(0, len(self.history))

        # The job now should be triggered with required label set
        fake_mr.approved = False
        fake_mr.labels = ['gateit']
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        # The job should not be triggered due to rejected label
        fake_mr.labels = ['gateit', 'do-not-merge']
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))

        # The job again should be triggered when rejected label is gone
        fake_mr.labels = ['gateit']
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

        # The job should not be triggered for closed change
        fake_mr.closeMergeRequest()
        self.fake_gitlab.emitEvent(
            fake_mr.getMergeRequestCommentedEvent('recheck')
        )
        self.waitUntilSettled()
        self.assertEqual(2, len(self.history))

        # The job again should be triggered after reopening the change
        fake_mr.reopenMergeRequest()
        self.fake_gitlab.emitEvent(fake_mr.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(3, len(self.history))

        # The job no longer should be triggered after the change was merged
        fake_mr.mergeMergeRequest()
        self.fake_gitlab.emitEvent(
            fake_mr.getMergeRequestCommentedEvent('recheck')
        )
        self.waitUntilSettled()
        self.assertEqual(3, len(self.history))

    @simple_layout('layouts/gitlab-label-add-remove.yaml', driver='gitlab')
    def test_label_add_remove(self):

        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project1', 'master', 'A')
        A.labels = ['removeme1', 'removeme2']

        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        self.assertEqual(1, len(self.history))
        self.assertEqual(set(A.labels), {'addme1', 'addme2'})

    @simple_layout('layouts/merging-gitlab.yaml', driver='gitlab')
    def test_merge_blocking_discussions(self):

        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project2', 'master', 'A')

        A.blocking_discussions_resolved = False
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        # canMerge is not validated
        self.assertEqual(0, len(self.history))

        A.blocking_discussions_resolved = True
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        # canMerge is validated
        self.assertEqual(1, len(self.history))

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test').result)
        self.assertEqual('merged', A.state)

    @simple_layout('layouts/merging-gitlab.yaml', driver='gitlab')
    def test_merge_action_in_independent(self):

        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project1', 'master', 'A')

        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()

        self.assertEqual(1, len(self.history))
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test').result)
        self.assertEqual('merged', A.state)

    @simple_layout('layouts/merging-gitlab.yaml', driver='gitlab')
    def test_merge_action_in_dependent(self):

        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project2', 'master', 'A')
        A.merge_status = 'cannot_be_merged'

        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        # canMerge is not validated
        self.assertEqual(0, len(self.history))

        # Set Merge request can be merged
        A.merge_status = 'can_be_merged'
        self.fake_gitlab.emitEvent(A.getMergeRequestUpdatedEvent())
        self.waitUntilSettled()
        # canMerge is validated
        self.assertEqual(1, len(self.history))

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test').result)
        self.assertEqual('merged', A.state)

    @simple_layout('layouts/merging-gitlab-squash-merge.yaml', driver='gitlab')
    def test_merge_squash(self):

        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project1', 'master', 'A')

        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        # canMerge is validated
        self.assertEqual(1, len(self.history))

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test').result)
        self.assertEqual('merged', A.state)
        self.assertTrue(A.squash_merge)

    @simple_layout('layouts/crd-gitlab.yaml', driver='gitlab')
    def test_crd_dependent(self):

        # Create a change in project3 that a project4 change will depend on
        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project3', 'master', 'A')
        A.approved = True

        # Create a change B that sets the dependency on A
        msg = "Depends-On: %s" % A.url
        B = self.fake_gitlab.openFakeMergeRequest(
            'org/project4', 'master', 'B', description=msg)

        # Emit B opened event
        event = B.getMergeRequestOpenedEvent()
        self.fake_gitlab.emitEvent(event)
        self.waitUntilSettled()

        # B cannot be merged (not approved)
        self.assertEqual(0, len(self.history))

        # Approve B
        B.approved = True
        # And send the event
        self.fake_gitlab.emitEvent(event)
        self.waitUntilSettled()

        # The changes for the job from project4 should include the project3
        # MR content
        changes = self.getJobFromHistory(
            'project4-test', 'org/project4').changes

        self.assertEqual(changes, "%s,%s %s,%s" % (A.number,
                                                   A.sha,
                                                   B.number,
                                                   B.sha))

        self.assertTrue(A.is_merged)
        self.assertTrue(B.is_merged)

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_timer_event(self):
        self.executor_server.hold_jobs_in_build = True
        self.commitConfigUpdate('org/common-config',
                                'layouts/timer-gitlab.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        time.sleep(2)
        self.waitUntilSettled()
        self.assertEqual(len(self.builds), 1)
        self.executor_server.hold_jobs_in_build = False
        # Stop queuing timer triggered jobs so that the assertions
        # below don't race against more jobs being queued.
        self.commitConfigUpdate('org/common-config',
                                'layouts/no-timer-gitlab.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()
        # If APScheduler is in mid-event when we remove the job, we
        # can end up with one more event firing, so give it an extra
        # second to settle.
        time.sleep(1)
        self.waitUntilSettled()
        self.executor_server.release()
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='project-bitrot', result='SUCCESS',
                 ref='refs/heads/master'),
        ], ordered=False)

    @simple_layout('layouts/crd-gitlab.yaml', driver='gitlab')
    def test_api_token(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        _, project = tenant.getProject('org/project1')

        project_git_url = self.fake_gitlab.real_getGitUrl(project)
        # cloneurl created from config 'server' should be used
        # without credentials
        self.assertEqual(f"{self.fake_gitlab._test_baseurl}/org/project1.git",
                         project_git_url)

    @simple_layout('layouts/crd-gitlab.yaml', driver='gitlab2')
    def test_api_token_cloneurl(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        _, project = tenant.getProject('org/project1')

        project_git_url = self.fake_gitlab2.real_getGitUrl(project)
        # cloneurl from config file should be used as it defines token name and
        # secret
        self.assertEqual("http://myusername:2222@gitlab/org/project1.git",
                         project_git_url)

    @simple_layout('layouts/crd-gitlab.yaml', driver='gitlab3')
    def test_api_token_name_cloneurl(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        _, project = tenant.getProject('org/project1')

        project_git_url = self.fake_gitlab3.real_getGitUrl(project)
        # cloneurl from config file should be used as it defines token name and
        # secret, even if token name and token secret are defined
        self.assertEqual("http://myusername:2222@gitlabthree/org/project1.git",
                         project_git_url)

    @simple_layout('layouts/crd-gitlab.yaml', driver='gitlab4')
    def test_api_token_name(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        _, project = tenant.getProject('org/project1')

        project_git_url = self.fake_gitlab4.real_getGitUrl(project)
        # cloneurl is not set, generate one from token name, token secret and
        # server
        self.assertEqual("http://tokenname4:444@localhost:"
                         f"{self.fake_gitlab4._test_web_server.port}"
                         "/org/project1.git",
                         project_git_url)

    @simple_layout('layouts/crd-gitlab.yaml', driver='gitlab5')
    def test_api_token_name_cloneurl_server(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        _, project = tenant.getProject('org/project1')

        project_git_url = self.fake_gitlab5.real_getGitUrl(project)
        # cloneurl defines a url, without credentials. As token name is
        # set, include token name and secret in cloneurl, 'server' is
        # overwritten
        self.assertEqual("http://tokenname5:555@gitlabfivvve/org/project1.git",
                         project_git_url)

    @simple_layout('layouts/files-gitlab.yaml', driver='gitlab')
    def test_changed_file_match_filter(self):
        path = os.path.join(self.upstream_root, 'org/project')
        base_sha = git.Repo(path).head.object.hexsha

        files = {'{:03d}.txt'.format(n): 'test' for n in range(300)}
        files["foobar-requires"] = "test"
        files["to-be-removed"] = "test"
        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project', 'master', 'A', files=files, base_sha=base_sha)

        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        # project-test1 and project-test2 should be run
        self.assertEqual(2, len(self.history))

    @simple_layout('layouts/files-gitlab.yaml', driver='gitlab')
    def test_changed_file_match_filter_refresh(self):
        # Test that a refresh of a MR doesn't overwrite the change
        # files list supplied by a zuul merger.
        self.hold_merge_jobs_in_queue = True
        path = os.path.join(self.upstream_root, 'org/project')
        base_sha = git.Repo(path).head.object.hexsha

        files = {}
        # This files causes project-test1 to run
        files["foobar-requires"] = "test"
        # File "to-be-removed" is not included, so -test2 should not run.
        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project', 'master', 'A', files=files, base_sha=base_sha)

        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()

        for x in self.merger_api.all():
            if x.job_type == 'fileschanges':
                self.merger_api.release(x)

        self.waitUntilSettled()
        self.fake_gitlab.emitEvent(A.getMergeRequestCommentedEvent(
            'test comment'))
        self.waitUntilSettled()
        self.hold_merge_jobs_in_queue = False
        self.merger_api.release()
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='project-test1', result='SUCCESS'),
        ])

    @simple_layout('layouts/files-gitlab.yaml', driver='gitlab')
    def test_changed_and_reverted_file_not_match_filter(self):
        path = os.path.join(self.upstream_root, 'org/project')
        base_sha = git.Repo(path).head.object.hexsha

        files = {'{:03d}.txt'.format(n): 'test' for n in range(300)}
        files["foobar-requires"] = "test"
        files["to-be-removed"] = "test"
        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project', 'master', 'A', files=files, base_sha=base_sha)
        A.addCommit(delete_files=['to-be-removed'])

        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()
        # Only project-test1 should be run, because the file to-be-removed
        # is reverted and not in changed files to trigger project-test2
        self.assertEqual(1, len(self.history))

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_update_change(self):
        # This tests the updateChange code path when no event is
        # present (ie, during a pipeline refresh).
        conn = self.scheds.first.sched.connections.connections['gitlab']
        conn.gl_client.get_mr_wait_factor = 0
        with self.fake_gitlab.enable_uncomplete_mr():
            A = self.fake_gitlab.openFakeMergeRequest(
                'org/project', 'master', 'A')
            change_key = ChangeKey(conn.connection_name, "org/project",
                                   "MergeRequest", str(A.number), A.sha)
            # diff_refs has not populated yet, so this call tests the
            # without-event and without-diff_refs path.
            change = conn.getChange(change_key)
            self.assertIsNone(change.commit_id, None)
            self.fake_gitlab.emitEvent(
                A.getMergeRequestOpenedEvent(), project='org/project')
            self.waitUntilSettled()
            # The above emits an event, so we should have populated
            # the change cache using the event info.  This tests the
            # with-event and without-diff_refs path.
            change = conn.getChange(change_key)
            self.assertEqual(change.commit_id, A.sha)

        # We don't test with event and with diff_refs here, since
        # having diff_refs supercedes using event data.  Also, any
        # other test in the class which does not set the uncomplete
        # flag will be testing that path.

        # Delete the change from the cache
        conn._change_cache.delete(change_key)
        # This will force the driver to fetch the change info without
        # an event.  This tests the without-event and with-diff_refs
        # path
        change = conn.getChange(change_key)
        self.assertEqual(change.commit_id, A.sha)

    @simple_layout("layouts/basic-gitlab.yaml", driver="gitlab")
    def test_get_project_branch_sha(self):
        # Exercise this method since it's only called from timer
        # triggers
        source = self.fake_gitlab.source
        project = source.getProject('org/project')
        self.assertIsNotNone(source.getProjectBranchSha(project, 'master'))

    @simple_layout('layouts/gitlab-pipeline-trigger-debug.yaml',
                   driver='gitlab')
    def test_gitlab_pipeline_trigger_debug(self):
        # Test that we get debug info for a pipeline debug trigger
        A = self.fake_gitlab.openFakeMergeRequest("org/project", "master", "A")
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()

        self.assertFalse(A.is_merged)
        self.assertEqual(1, len(A.notes))
        self.assertIn('Debug information:',
                      A.notes[0]['body'])

        # Test that we get a reason from canMerge.
        B = self.fake_gitlab.openFakeMergeRequest("org/project", "master", "B")
        B.blocking_discussions_resolved = False
        self.fake_gitlab.emitEvent(B.getMergeRequestLabeledEvent(
            add_labels=('gateit', )))
        self.waitUntilSettled()

        self.assertFalse(B.is_merged)
        self.assertEqual(1, len(B.notes))
        self.assertIn("can not be merged due to: blocking discussions",
                      B.notes[0]['body'])

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_matched_branch_event(self):
        """Test merge request targeting branch that matches pipeline config"""
        A = self.fake_gitlab.openFakeMergeRequest('org/project', 'master', 'A')
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test1').result)
        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test2').result)
        self.assertEqual(2, len(self.history))

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_unmatched_branch_event(self):
        """Test merge request targeting branch that doesn't match pipeline"""
        # Create an unmatched branch
        self.create_branch('org/project', 'unmatched_branch')

        # Create a merge request targeting the unmatched branch
        B = self.fake_gitlab.openFakeMergeRequest(
            'org/project', 'unmatched_branch', 'B')
        self.fake_gitlab.emitEvent(B.getMergeRequestOpenedEvent())
        self.waitUntilSettled()

        # No jobs should run for unmatched branch
        self.assertEqual(0, len(self.history))


class TestGitlabUnprotectedBranches(ZuulTestCase):
    config_file = 'zuul-gitlab-driver.conf'
    tenant_config_file = 'config/unprotected-branches-gitlab/main.yaml'

    @skipIfMultiScheduler()
    # This test is failing with multiple schedulers depending on which
    # scheduler did the tenant reconfiguration first. As the
    # assertions are all done on the objects from scheduler-0 they
    # will fail if scheduler-1 did the reconfig first.
    # To make this work with multiple schedulers, we might want to wait
    # until all schedulers completed their tenant reconfiguration.
    def test_unprotected_branches(self):
        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        layout = tenant.layout

        # project1 should have parsed master
        self.assertIn('project1-job', layout.jobs)

        # project2 should have no parsed branch
        self.assertNotIn('project2-job', layout.jobs)

        # now enable branch protection and trigger reload
        self.fake_gitlab.protectBranch('org', 'project2', 'master')
        pevent = self.fake_gitlab.getPushEvent(project='org/project2')
        self.fake_gitlab.emitEvent(pevent)
        self.waitUntilSettled()

        tenant = self.scheds.first.sched.abide.tenants.get('tenant-one')
        layout = tenant.layout

        # project1 and project2 should have parsed master now
        self.assertIn('project1-job', layout.jobs)
        self.assertIn('project2-job', layout.jobs)

    def test_filtered_branches_in_build(self):
        """
        Tests unprotected branches are filtered in builds if excluded
        """
        self.executor_server.keep_jobdir = True

        # Enable branch protection on org/project2@master
        self.create_branch('org/project2', 'feat-x')
        self.fake_gitlab.protectBranch('org', 'project2', 'master',
                                       protected=True)

        # Enable branch protection on org/project3@stable. We'll use a MR on
        # this branch as a depends-on to validate that the stable branch
        # which is not protected in org/project2 is not filtered out.
        self.create_branch('org/project3', 'stable')
        self.fake_gitlab.protectBranch('org', 'project3', 'stable',
                                       protected=True)

        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        A = self.fake_gitlab.openFakeMergeRequest('org/project3', 'stable',
                                                  'A')
        msg = "Depends-On: %s" % A.url
        B = self.fake_gitlab.openFakeMergeRequest('org/project2', 'master',
                                                  'B', description=msg)

        self.fake_gitlab.emitEvent(B.getMergeRequestOpenedEvent())
        self.waitUntilSettled()

        build = self.history[0]
        path = os.path.join(
            build.jobdir.src_root, 'gitlab', 'org/project2')
        build_repo = git.Repo(path)
        branches = [x.name for x in build_repo.branches]
        self.assertNotIn('feat-x', branches)

        self.assertHistory([
            dict(name='used-job', result='SUCCESS',
                 changes="%s,%s %s,%s" % (A.number, A.sha,
                                          B.number, B.sha)),
        ])

    def test_unfiltered_branches_in_build(self):
        """
        Tests unprotected branches are not filtered in builds if not excluded
        """
        self.executor_server.keep_jobdir = True

        # Enable branch protection on org/project1@master
        self.create_branch('org/project1', 'feat-x')
        self.fake_gitlab.protectBranch('org', 'project1', 'master',
                                       protected=True)
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        A = self.fake_gitlab.openFakeMergeRequest('org/project1', 'master',
                                                  'A')
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()

        build = self.history[0]
        path = os.path.join(
            build.jobdir.src_root, 'gitlab', 'org/project1')
        build_repo = git.Repo(path)
        branches = [x.name for x in build_repo.branches]
        self.assertIn('feat-x', branches)

        self.assertHistory([
            dict(name='project-test', result='SUCCESS',
                 changes="%s,%s" % (A.number, A.sha)),
        ])

    def test_unprotected_push(self):
        """Test that unprotected pushes don't cause tenant reconfigurations"""

        # Prepare repo with an initial commit
        A = self.fake_gitlab.openFakeMergeRequest('org/project2', 'master',
                                                  'A')

        zuul_yaml = [
            {'job': {
                'name': 'used-job2',
                'run': 'playbooks/used-job.yaml'
            }},
            {'project': {
                'check': {
                    'jobs': [
                        'used-job2'
                    ]
                }
            }}
        ]

        A.addCommit({'zuul.yaml': yaml.dump(zuul_yaml)})
        A.mergeMergeRequest()

        # Do a push on top of A
        pevent = self.fake_gitlab.getPushEvent(project='org/project2',
                                               before=A.sha,
                                               branch='refs/heads/master')

        # record previous tenant reconfiguration time, which may not be set
        old = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)
        self.waitUntilSettled()

        self.fake_gitlab.emitEvent(pevent)
        self.waitUntilSettled()
        new = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)

        # We don't expect a reconfiguration because the push was to an
        # unprotected branch
        self.assertEqual(old, new)

        # now enable branch protection and trigger the push event again
        self.fake_gitlab.protectBranch('org', 'project2', 'master')

        self.fake_gitlab.emitEvent(pevent)
        self.waitUntilSettled()
        new = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)

        # We now expect that zuul reconfigured itself
        self.assertLess(old, new)

    def test_protected_branch_delete(self):
        """Test that protected branch deletes trigger a tenant reconfig"""

        # Prepare repo with an initial commit and enable branch protection
        self.fake_gitlab.protectBranch('org', 'project2', 'master')
        self.fake_gitlab.emitEvent(
            self.fake_gitlab.getPushEvent(
                project='org/project2', branch='refs/heads/master'))

        A = self.fake_gitlab.openFakeMergeRequest('org/project2', 'master',
                                                  'A')
        A.mergeMergeRequest()

        # add a spare branch so that the project is not empty after master gets
        # deleted.
        self.create_branch('org/project2', 'feat-x')
        self.fake_gitlab.protectBranch('org', 'project2', 'feat-x',
                                       protected=False)
        self.fake_gitlab.emitEvent(
            self.fake_gitlab.getPushEvent(
                project='org/project2', branch='refs/heads/feat-x'))
        self.waitUntilSettled()

        # record previous tenant reconfiguration time, which may not be set
        old = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)
        self.waitUntilSettled()

        # Delete the branch
        self.fake_gitlab.deleteBranch('org', 'project2', 'master')

        pevent = self.fake_gitlab.getPushEvent(project='org/project2',
                                               before=A.sha,
                                               after='0' * 40,
                                               branch='refs/heads/master')

        self.fake_gitlab.emitEvent(pevent)
        self.waitUntilSettled()
        new = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)

        # We now expect that zuul reconfigured itself as we deleted a protected
        # branch
        self.assertLess(old, new)

    # This test verifies that a PR is considered in case it was created for
    # a branch just has been set to protected before a tenant reconfiguration
    # took place.
    def test_reconfigure_on_pr_to_new_protected_branch(self):
        self.create_branch('org/project2', 'release')
        self.create_branch('org/project2', 'feature')

        self.fake_gitlab.protectBranch('org', 'project2', 'master')
        self.fake_gitlab.protectBranch('org', 'project2', 'release',
                                       protected=False)
        self.fake_gitlab.protectBranch('org', 'project2', 'feature',
                                       protected=False)

        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        self.fake_gitlab.protectBranch('org', 'project2', 'release')

        self.executor_server.hold_jobs_in_build = True

        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project2', 'release', 'A')
        self.fake_gitlab.emitEvent(A.getMergeRequestOpenedEvent())
        self.waitUntilSettled()

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('used-job').result)

        job = self.getJobFromHistory('used-job')
        zuulvars = job.parameters['zuul']
        self.assertEqual(str(A.number), zuulvars['change'])
        self.assertEqual(str(A.sha), zuulvars['patchset'])
        self.assertEqual('release', zuulvars['branch'])

        self.assertEqual(1, len(self.history))

    def _test_push_event_reconfigure(self, project, branch,
                                     expect_reconfigure=False,
                                     old_sha=None, new_sha=None,
                                     modified_files=None,
                                     removed_files=None):
        pevent = self.fake_gitlab.getPushEvent(
            project=project,
            branch='refs/heads/%s' % branch,
            before=old_sha,
            after=new_sha)

        # record previous tenant reconfiguration time, which may not be set
        old = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)
        self.waitUntilSettled()

        self.fake_gitlab.emitEvent(pevent)
        self.waitUntilSettled()
        new = self.scheds.first.sched.tenant_layout_state.get(
            'tenant-one', EMPTY_LAYOUT_STATE)

        if expect_reconfigure:
            # New timestamp should be greater than the old timestamp
            self.assertLess(old, new)
        else:
            # Timestamps should be equal as no reconfiguration shall happen
            self.assertEqual(old, new)

    def test_push_event_reconfigure_complex_branch(self):

        branch = 'feature/somefeature'
        project = 'org/project2'

        # prepare an existing branch
        self.create_branch(project, branch)

        self.fake_gitlab.protectBranch(*project.split('/'), branch,
                                       protected=False)

        self.fake_gitlab.emitEvent(
            self.fake_gitlab.getPushEvent(
                project,
                branch='refs/heads/%s' % branch))
        self.waitUntilSettled()

        A = self.fake_gitlab.openFakeMergeRequest(project, branch, 'A')
        old_sha = A.sha
        A.mergeMergeRequest()
        new_sha = random_sha1()

        # branch is not protected, no reconfiguration even if config file
        self._test_push_event_reconfigure(project, branch,
                                          expect_reconfigure=False,
                                          old_sha=old_sha,
                                          new_sha=new_sha,
                                          modified_files=['zuul.yaml'])

        # branch is not protected: no reconfiguration
        self.fake_gitlab.deleteBranch(*project.split('/'), branch)

        self._test_push_event_reconfigure(project, branch,
                                          expect_reconfigure=False,
                                          old_sha=new_sha,
                                          new_sha='0' * 40,
                                          removed_files=['zuul.yaml'])


class TestGitlabDriverNoPool(ZuulTestCase):
    config_file = 'zuul-gitlab-driver-no-pool.conf'

    @simple_layout('layouts/basic-gitlab.yaml', driver='gitlab')
    def test_merge_request_opened(self):

        description = "This is the\nMR description."
        A = self.fake_gitlab.openFakeMergeRequest(
            'org/project', 'master', 'A', description=description)
        self.fake_gitlab.emitEvent(
            A.getMergeRequestOpenedEvent(), project='org/project')
        self.waitUntilSettled()

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test1').result)

        self.assertEqual('SUCCESS',
                         self.getJobFromHistory('project-test2').result)

        job = self.getJobFromHistory('project-test2')
        zuulvars = job.parameters['zuul']
        self.assertEqual(str(A.number), zuulvars['change'])
        self.assertEqual(str(A.sha), zuulvars['patchset'])
        self.assertEqual('master', zuulvars['branch'])
        self.assertEqual(f'{self.fake_gitlab._test_baseurl}/'
                         'org/project/merge_requests/1',
                         zuulvars['items'][0]['change_url'])
        self.assertEqual(zuulvars["message"], strings.b64encode(description))
        self.assertEqual(2, len(self.history))
        self.assertEqual(2, len(A.notes))
        self.assertEqual(
            A.notes[0]['body'], "Starting check jobs.")
        self.assertThat(
            A.notes[1]['body'],
            MatchesRegex(r'.*project-test1.*SUCCESS.*', re.DOTALL))
        self.assertThat(
            A.notes[1]['body'],
            MatchesRegex(r'.*project-test2.*SUCCESS.*', re.DOTALL))
        self.assertTrue(A.approved)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import os
import tempfile
import time

from tests.base import BaseTestCase

from zuul.executor.server import DiskAccountant


class FakeExecutor(object):
    def __init__(self):
        self.stopped_jobs = set()
        self.used = {}

    def stopJobByJobDir(self, jobdir):
        self.stopped_jobs.add(jobdir)

    def usage(self, dirname, used):
        self.used[dirname] = used


class TestDiskAccountant(BaseTestCase):
    def setUp(self):
        super(TestDiskAccountant, self).setUp()
        self.useFixture(fixtures.NestedTempfile())

    def test_disk_accountant(self):
        jobs_dir = tempfile.mkdtemp(
            dir=os.environ.get("ZUUL_TEST_ROOT", None))
        cache_dir = tempfile.mkdtemp()
        executor_server = FakeExecutor()
        da = DiskAccountant(jobs_dir, 1, executor_server.stopJobByJobDir,
                            cache_dir)
        da.start()

        try:
            jobdir = os.path.join(jobs_dir, '012345')
            os.mkdir(jobdir)
            testfile = os.path.join(jobdir, 'tfile')
            with open(testfile, 'w') as tf:
                tf.write(2 * 1024 * 1024 * '.')
                tf.flush()
                os.fsync(tf.fileno())

            # da should catch over-limit dir within 5 seconds
            for i in range(0, 50):
                if jobdir in executor_server.stopped_jobs:
                    break
                time.sleep(0.1)
            self.assertEqual(set([jobdir]), executor_server.stopped_jobs)
        finally:
            da.stop()
        self.assertFalse(da.thread.is_alive())

    def test_disk_accountant_no_limit(self):
        jobs_dir = tempfile.mkdtemp(
            dir=os.environ.get("ZUUL_TEST_ROOT", None))
        cache_dir = tempfile.mkdtemp()
        executor_server = FakeExecutor()
        da = DiskAccountant(jobs_dir, -1, executor_server.stopJobByJobDir,
                            cache_dir)
        da.start()
        self.assertFalse(da.running)
        da.stop()
        self.assertFalse(da.running)

    def test_cache_hard_links(self):
        root_dir = tempfile.mkdtemp(
            dir=os.environ.get("ZUUL_TEST_ROOT", None))
        jobs_dir = os.path.join(root_dir, 'jobs')
        os.mkdir(jobs_dir)
        cache_dir = os.path.join(root_dir, 'cache')
        os.mkdir(cache_dir)

        executor_server = FakeExecutor()
        da = DiskAccountant(jobs_dir, 1, executor_server.stopJobByJobDir,
                            cache_dir, executor_server.usage)
        da.start()
        self.addCleanup(da.stop)

        jobdir = os.path.join(jobs_dir, '012345')
        os.mkdir(jobdir)

        repo_dir = os.path.join(cache_dir, 'a.repo')
        os.mkdir(repo_dir)
        source_file = os.path.join(repo_dir, 'big_file')
        with open(source_file, 'w') as tf:
            tf.write(2 * 1024 * 1024 * '.')
        dest_link = os.path.join(jobdir, 'big_file')
        os.link(source_file, dest_link)

        # da should _not_ count this file. Wait for 5s to get noticed
        for i in range(0, 50):
            if jobdir in executor_server.used:
                break
            time.sleep(0.1)
        self.assertEqual(set(), executor_server.stopped_jobs)
        self.assertIn(jobdir, executor_server.used)
        self.assertTrue(executor_server.used[jobdir] <= 1)



# Copyright 2015 Red Hat, Inc.
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
This module defines classes used in matching changes based on job
configuration.
"""

import json

from zuul.lib.re2util import ZuulRegex

COMMIT_MSG = '/COMMIT_MSG'


class AbstractChangeMatcher(object):
    """An abstract class that matches change attributes against regular
    expressions

    :arg ZuulRegex regex: A Zuul regular expression object to match

    """

    def __init__(self, regex):
        self._regex = regex.pattern
        self.regex = regex

    def matches(self, change):
        """Return a boolean indication of whether change matches
        implementation-specific criteria.
        """
        raise NotImplementedError()

    def copy(self):
        return self.__class__(self.regex)

    def __deepcopy__(self, memo):
        return self.copy()

    def __eq__(self, other):
        return (self.__class__ == other.__class__ and
                self.regex == other.regex)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(json.dumps(self.regex.toDict(), sort_keys=True))

    def __str__(self):
        return '{%s:%s}' % (self.__class__.__name__, self._regex)

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self._regex)


class ProjectMatcher(AbstractChangeMatcher):

    def matches(self, change):
        return self.regex.match(str(change.project))


class BranchMatcher(AbstractChangeMatcher):
    exactmatch = False

    def matches(self, change):
        if hasattr(change, 'branch'):
            # an implied branch matcher must do a fullmatch to work correctly
            if self.exactmatch:
                if self._regex == change.branch:
                    return True
            else:
                if self.regex.match(change.branch):
                    return True
            return False
        if self.regex.match(change.ref):
            return True
        if hasattr(change, 'containing_branches'):
            for branch in change.containing_branches:
                if self.exactmatch:
                    if self._regex == branch:
                        return True
                else:
                    if self.regex.match(branch):
                        return True
        return False

    def serialize(self):
        return {
            "implied": self.exactmatch,
            "regex": self.regex.serialize(),
        }

    @classmethod
    def deserialize(cls, data):
        regex = ZuulRegex.deserialize(data['regex'])
        o = cls(regex)
        return o


class ImpliedBranchMatcher(BranchMatcher):
    exactmatch = True


class FileMatcher(AbstractChangeMatcher):

    pass


class AbstractMatcherCollection(AbstractChangeMatcher):

    def __init__(self, matchers):
        self.matchers = matchers

    def __eq__(self, other):
        return str(self) == str(other)

    def __str__(self):
        return '{%s:%s}' % (self.__class__.__name__,
                            ','.join([str(x) for x in self.matchers]))

    def __repr__(self):
        return '<%s [%s]>' % (self.__class__.__name__,
                              ', '.join([repr(x) for x in self.matchers]))

    def copy(self):
        return self.__class__(self.matchers[:])


class AbstractMatchFiles(AbstractMatcherCollection):

    @property
    def regexes(self):
        for matcher in self.matchers:
            yield matcher.regex


class MatchAllFiles(AbstractMatchFiles):

    def matches(self, change):
        # NOTE(yoctozepto): make irrelevant files matcher match when
        # there are no files to check - return False (NB: reversed)
        if not hasattr(change, 'files'):
            return False
        files = [c for c in change.files if c != COMMIT_MSG]
        if not files:
            return False
        for file_ in files:
            matched_file = False
            for regex in self.regexes:
                if regex.match(file_):
                    matched_file = True
                    break
            if not matched_file:
                return False
        return True


class MatchAnyFiles(AbstractMatchFiles):

    def matches(self, change):
        # NOTE(yoctozepto): make files matcher match when
        # there are no files to check - return True
        if not hasattr(change, 'files'):
            return True
        files = [c for c in change.files if c != COMMIT_MSG]
        if not files:
            return True
        for file_ in files:
            for regex in self.regexes:
                if regex.match(file_):
                    return True
        return False


class MatchAll(AbstractMatcherCollection):

    def matches(self, change):
        for matcher in self.matchers:
            if not matcher.matches(change):
                return False
        return True


class MatchAny(AbstractMatcherCollection):

    def matches(self, change):
        for matcher in self.matchers:
            if matcher.matches(change):
                return True
        return False



#!/usr/bin/env python3

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import base64
import json
import math
import os
import re
import subprocess
import sys
import tempfile
import textwrap
import ssl

# we to import Request and urlopen differently for python 2 and 3
try:
    from urllib.request import Request
    from urllib.request import urlopen
    from urllib.parse import urlparse
except ImportError:
    from urllib2 import Request
    from urllib2 import urlopen
    from urlparse import urlparse

DESCRIPTION = """Encrypt a secret for Zuul.

This program fetches a project-specific public key from a Zuul server and
uses that to encrypt a secret.  The only pre-requisite is an installed
OpenSSL binary.
"""


def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('url',
                        help="The base URL of the zuul server.  "
                        "E.g., https://zuul.example.com/ or path"
                        " to project public key file. E.g.,"
                        " file:///path/to/key.pub")
    parser.add_argument('project', default=None, nargs="?",
                        help="The name of the project. Required when using"
                        " the Zuul API to fetch the public key.")
    parser.add_argument('--tenant',
                        default=None,
                        help="The name of the Zuul tenant.  This may be "
                        "required in a multi-tenant environment.")
    parser.add_argument('--strip', default=None,
                        help='Unused, kept for backward compatibility.')
    parser.add_argument('--no-strip', action='store_true', default=False,
                        help="Do not strip whitespace from beginning or "
                        "end of input.")
    parser.add_argument('--infile',
                        default=None,
                        help="A filename whose contents will be encrypted.  "
                        "If not supplied, the value will be read from "
                        "standard input.")
    parser.add_argument('--outfile',
                        default=None,
                        help="A filename to which the encrypted value will be "
                        "written.  If not supplied, the value will be written "
                        "to standard output.")
    parser.add_argument('--insecure', action='store_true', default=False,
                        help="Do not verify remote certificate")
    args = parser.parse_args()

    # We should not use unencrypted connections for retrieving the public key.
    # Otherwise our secret can be compromised. The schemes file and https are
    # considered safe.
    url = urlparse(args.url)
    if url.scheme not in ('file', 'https'):
        sys.stderr.write("WARNING: Retrieving encryption key via an "
                         "unencrypted connection. Your secret may get "
                         "compromised.\n")

    ssl_ctx = None
    if url.scheme == 'file':
        req = Request(args.url)
    else:
        if args.insecure:
            ssl_ctx = ssl.create_default_context()
            ssl_ctx.check_hostname = False
            ssl_ctx.verify_mode = ssl.CERT_NONE

        # Check if tenant is white label
        req = Request("%s/api/info" % (args.url.rstrip('/'),))
        info = json.loads(urlopen(req, context=ssl_ctx).read().decode('utf8'))

        api_tenant = info.get('info', {}).get('tenant')
        if not api_tenant and not args.tenant:
            print("Error: the --tenant argument is required")
            exit(1)

        if api_tenant:
            req = Request("%s/api/key/%s.pub" % (
                args.url.rstrip('/'), args.project))
        else:
            req = Request("%s/api/tenant/%s/key/%s.pub" % (
                args.url.rstrip('/'), args.tenant, args.project))
    try:
        pubkey = urlopen(req, context=ssl_ctx)
    except Exception:
        sys.stderr.write(
            "ERROR: Couldn't retrieve project key via %s\n" % req.full_url)
        raise

    if args.infile:
        with open(args.infile) as f:
            plaintext = f.read()
    else:
        plaintext = sys.stdin.read()

    plaintext = plaintext.encode("utf-8")
    if not args.no_strip:
        plaintext = plaintext.strip()

    pubkey_file = tempfile.NamedTemporaryFile(delete=False)
    try:
        pubkey_file.write(pubkey.read())
        pubkey_file.close()

        p = subprocess.Popen(['openssl', 'rsa', '-text',
                              '-pubin', '-in',
                              pubkey_file.name],
                             stdout=subprocess.PIPE)
        (stdout, stderr) = p.communicate()
        if p.returncode != 0:
            raise Exception("Return code %s from openssl" % p.returncode)
        output = stdout.decode('utf-8')
        openssl_version = subprocess.check_output(
            ['openssl', 'version']).split()[1]
        if openssl_version.startswith(b'0.'):
            key_length_re = r'^Modulus \((?P<key_length>\d+) bit\):$'
        else:
            key_length_re = r'^(|RSA )Public-Key: \((?P<key_length>\d+) bit\)$'
        m = re.match(key_length_re, output, re.MULTILINE)
        nbits = int(m.group('key_length'))
        nbytes = int(nbits / 8)
        max_bytes = nbytes - 42  # PKCS1-OAEP overhead
        chunks = int(math.ceil(float(len(plaintext)) / max_bytes))

        ciphertext_chunks = []

        print("Public key length: {} bits ({} bytes)".format(nbits, nbytes))
        print("Max plaintext length per chunk: {} bytes".format(max_bytes))
        print("Input plaintext length: {} bytes".format(len(plaintext)))
        print("Number of chunks: {}".format(chunks))

        for count in range(chunks):
            chunk = plaintext[int(count * max_bytes):
                              int((count + 1) * max_bytes)]
            p = subprocess.Popen(['openssl', 'rsautl', '-encrypt',
                                  '-oaep', '-pubin', '-inkey',
                                  pubkey_file.name],
                                 stdin=subprocess.PIPE,
                                 stdout=subprocess.PIPE)
            (stdout, stderr) = p.communicate(chunk)
            if p.returncode != 0:
                raise Exception("Return code %s from openssl" % p.returncode)
            ciphertext_chunks.append(base64.b64encode(stdout).decode('utf-8'))

    finally:
        os.unlink(pubkey_file.name)

    output = textwrap.dedent(
        '''
        - secret:
            name: <name>
            data:
              <fieldname>: !encrypted/pkcs1-oaep
        ''')

    twrap = textwrap.TextWrapper(width=79,
                                 initial_indent=' ' * 8,
                                 subsequent_indent=' ' * 10)
    for chunk in ciphertext_chunks:
        chunk = twrap.fill('- ' + chunk)
        output += chunk + '\n'

    if args.outfile:
        with open(args.outfile, "w") as f:
            f.write(output)
    else:
        print(output)


if __name__ == '__main__':
    print(
        "This script is deprecated. Use `zuul-client encrypt` instead. "
        "Please refer to https://zuul-ci.org/docs/zuul-client/ "
        "for more details on how to use zuul-client."
    )
    main()
